// Title:   DefiningSurfaces.cpp
// Created: Thu Sep 25 15:16:24 2003
// Authors: Yong Joo Kil (modified by)
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "DefiningSurfaces.h"
#ifdef _WIN32
	// turn off std::map warnings
	//# pragma warning(push)
	# pragma warning(disable:4786) 
#endif

#include "MovingLeastSquares.h"

#include <vector>
#include <qdatetime.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_errno.h>
#include <qapplication.h>

#include "my_defs.h"
#include "UndoRedoSpecialisations.h"

#define NEIGHBOURHOOD_SIZE 10

#include "KdTreeExtra.h"
#include "PointNormalSurface.h"
#include "../../../../Utilities/UserMessages/src/StatusBar.h"

#include "NewOverload.h"

DefiningSurfaces::DefiningSurfaces() : NormalAction() {
	_name = "moving least squares smoother";
	_selection_index = 0;

	_sp.neighbourhoodSize = NEIGHBOURHOOD_SIZE;
}

void DefiningSurfaces::setParameters(int nhood_size, float gauss_factor, float alpha, bool useplane, int degree, 
			int xextremalSurfaceMethod,
			bool xweightbyRadiusBox, float xradiusScale,
			float xmahalanobisScale,
			int	xbrentMaxIter, float xbrentTol, float xbrentSearchRange,
			bool xprojectSelectionOnly) {

	_sp.neighbourhoodSize = nhood_size;

	_alpha = alpha;
	_useplane = useplane;
	_degree = degree;

	_sp.gaussFactor = gauss_factor;
	_sp.gaussFactorSquared = gauss_factor * gauss_factor;

	_sp.extremalSurfaceMethod = xextremalSurfaceMethod;
	_sp.useRadiusAsWeight = xweightbyRadiusBox;
	_sp.radiusScale = xradiusScale;
	
	_sp.mahalanobisScale = xmahalanobisScale;

	_sp.brentMaxIter = xbrentMaxIter;
	_sp.brentTol = xbrentTol;
	_sp.brentSearchRange = xbrentSearchRange;

	_projectSelectionOnly = xprojectSelectionOnly;
}

/**
 * this version uses the movingleastsquares.cpp implementation
 */

UndoAction* DefiningSurfaces::executeUndo() {
	QApplication::setOverrideCursor (Qt::waitCursor);

	Selector *selector;
	std::vector<unsigned int> *selection;
	std::vector<SurfelInterface*> old_surfels;
	std::vector<Vector3D> old_positions;
	std::vector<Vector3D> old_normals;
	std::vector<Vector3D>			*surface_positions = 0;
	std::vector<Vector3D>			*surface_normals = 0;
	std::vector<SurfelInterface*>	*surface_surfels = 0;
	std::vector<float>				*surface_radii = 0;
	Vector3D curr_pos;
	Vector3D normal;
	Vector3D new_pos;
	Vector3D new_normal;
	StatusBar			* statusBar = StatusBar::getInstance();

	QTime time;

	statusBar->showMessage("reading surfels ...", -1);

	selector = Selector::getInstance();
	selector->buildTree();
	selection = selector->getToProjectSurfelIndicesStdVec(Selector::intToFlag(_selection_index), _projectSelectionOnly);
	if (selection->size() == 0) {
		QApplication::restoreOverrideCursor();
		_MCdelete((unsigned long)selection);
		delete selection;
		return 0;
	}

	// get surface informations
	selector->getSurfacePositionNormalSurfelRadiusWorldStdVec( &surface_positions, &surface_normals, &surface_surfels, &surface_radii);

	PointNormalSurface PNS(_sp);
	PNS.setPositionsNormals(*surface_positions, *surface_normals, *surface_surfels, *surface_radii);

	statusBar->resetProgress();
	statusBar->showMessage("Projecting ...", -1);
	// do projections
	unsigned int nSurfels = selection->size();
	time.start();
	for (unsigned int i=0; i<nSurfels; i++) {
		curr_pos = selector->getSurfelWorldPosition((*selection)[i]);
		PNS.setSource(curr_pos);
		PNS.doPlaneProjectionEnergyRealParticle();
		
		if (_useplane) {
			new_pos = PNS.getPlanePosition() * _alpha + curr_pos * (1.0 - _alpha);
			
			SurfelInterface *surfel = selector->getSurfel((*selection)[i]);
			old_surfels.push_back(surfel);
			old_positions.push_back(surfel->getPosition());
			old_normals.push_back(surfel->getNormal());
			
			selector->setSurfelWorldPosition((*selection)[i], new_pos, false);
			selector->setSurfelWorldNormal((*selection)[i], PNS.getPlaneNormal(),false);
		} else {
			PNS.doPolynomialProjection(_degree);
			new_pos = PNS.getPolyPosition() * _alpha + curr_pos * (1.0 - _alpha);
		
			SurfelInterface *surfel = selector->getSurfel((*selection)[i]);
			old_surfels.push_back(surfel);
			old_positions.push_back(surfel->getPosition());
			old_normals.push_back(surfel->getNormal());

			selector->setSurfelWorldPosition((*selection)[i], new_pos, false);
			// using normal of polyfit
			selector->setSurfelWorldNormal((*selection)[i], PNS.getPlaneNormal(),false);
		}
		if (i % (nSurfels / 100 > 0 ? nSurfels / 100 : 1) == 0) {
			statusBar->setProgress ((int)((float)i / (float)nSurfels * 100.0f), false);
		}
	}		
	statusBar->setProgress(100);

	Scene::getInstance()->emitObjectModifiedSignal();
	Scene::getInstance()->emitRenderFrameSignal();

	qDebug("%d msecs calculated",time.elapsed());
	QApplication::restoreOverrideCursor();

	char str[128];
	sprintf(str, "%d msecs calculated ",time.elapsed());
	statusBar->showMessage (str, -1);

	PositionNormalUndo *undo = new PositionNormalUndo(old_surfels, old_positions, old_normals);
	_MCnew( (unsigned long)undo, sizeof(undo), __FILE__, __LINE__);
	
	_MCdelete((unsigned long)selection);
	delete selection;
	_MCdelete((unsigned long)surface_positions);
	delete surface_positions;
	_MCdelete((unsigned long)surface_normals);
	delete surface_normals;
	_MCdelete((unsigned long)surface_surfels);
	delete surface_surfels;
	_MCdelete((unsigned long)surface_radii);
	delete surface_radii;

	old_positions.clear();
	old_normals.clear();
	old_surfels.clear();
	return undo;
}

void DefiningSurfaces::execute() {
	UndoAction *undo = executeUndo();
	delete undo;
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
