// Title:   Energies.cpp
// Created: May 28 2004
// Authors: Yong Joo Kil
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//


#include <gsl/gsl_matrix.h>
#include <gsl/gsl_eigen.h>
#include "../../../../DataTypes/src/Vector3D.h"
#include "../../../../Core/DataStructures/src/NeighbourHood.h"
#include "DefiningSurfaceTypes.h"
#include "Energies.h"
#include "Selector.h"

#include "NewOverload.h"

float energyDummy(KdTreeExtra *kdtree, const SurfaceParams & sp, const Vector3D & source, const Vector3D & n) {
	return 0;
}

bool optDirDummy(KdTreeExtra *kdtree, const SurfaceParams & sp, const Vector3D & x, Vector3D * xnormal) {
	(*xnormal) = Vector3D(1,1,1);  //some unexpected normal direction show that we know this function was called.
	xnormal->normalize();
	return true;
}

float energyMahalanobis(KdTreeExtra *kdtree, const SurfaceParams & sp, const Vector3D & source, const Vector3D & n) {
	int i, surfel_len = kdtree->getNOfFoundNeighbours();
	Vector3D pos, norm, projx, u, v, vec;
	float weight = 0;
	float dist_squared, dist2plane_squared, rEnergy = 0, totalweight = 0;
	float min_dist = FLT_MAX, min_dist2plane_squared = 0;
	float h1 = 1, h2 = 1, h3 = 1;
	h1 = sp.mahalanobisScale;
	float param = sp.gaussFactorSquared;

	for (i=0; i<surfel_len; i++) {
		int index = kdtree->getNeighbourPositionIndex(i);
		const Vector3D & pi = kdtree->getNeighbourPositionByIndex(index);
		const Vector3D & ni = kdtree->getNeighbourNormalByIndex(index);
		if (sp.useRadiusAsWeight) {
			param = kdtree->getNeighbourRadiusByIndex(index);
		}
		ni.getTwoOrthogonals(u,v); u.normalize(); v.normalize();
		vec = source - pi;

		float a, b, c;
		a = Vector3D::dotProduct(vec, ni);
		b = Vector3D::dotProduct(vec, u);
		c = Vector3D::dotProduct(vec, v);
		dist2plane_squared = h1*a*a + h2*b*b + h3*c*c;

		// get weights
		dist_squared = Vector3D::squaredDistance(pi, source);
		weight = exp(- dist_squared / param);
		totalweight += weight;

		// Total distance sum
		rEnergy += dist2plane_squared * weight;
		if (dist_squared < min_dist) {
			min_dist = dist_squared;
			min_dist2plane_squared = dist2plane_squared;
		}

		//since we are using exp fall off and the distances are ordered,
		//following weights will also have all zeros.
		if (weight == 0)
			break;
	}

	if (totalweight <= 0) {
		return (float) min_dist2plane_squared;
	}
	rEnergy /= totalweight;
	return (float) rEnergy;
}

// Sum of <a, pi - x>^2 * Delta(x,pi)
float energyMLS(KdTreeExtra *kdtree, const SurfaceParams & sp, const Vector3D & x, const Vector3D & a) {
	int i, surfel_len = kdtree->getNOfFoundNeighbours();
	float weight = 0;
	float dist_squared, rEnergy = 0, totalweight = 0;
	float param = sp.gaussFactorSquared;
	Vector3D v;

	for (i=0; i<surfel_len; i++) {
		int index = kdtree->getNeighbourPositionIndex(i);
		const Vector3D & pi = kdtree->getNeighbourPositionByIndex(index);
		const Vector3D & ni = kdtree->getNeighbourNormalByIndex(index);
		if (sp.useRadiusAsWeight) {
			param = kdtree->getNeighbourRadiusByIndex(index);
		}

		//<a,pi-x>^2
		v = pi - x;
		float a_dot_v =	Vector3D::dotProduct(a,v);
		float dist2plane_squared = a_dot_v * a_dot_v;

		// Delta(x,pi) weight
		dist_squared = v.getSquaredLength();
		weight = exp(- dist_squared / param);
		totalweight += weight;

		// Total distance sum
		rEnergy += dist2plane_squared * weight;

		//since we are using exp fall off and the distances are ordered,
		//i + 1 weights will also have all zeros.
		if (weight == 0)
			break;
	}

	return (float) rEnergy;
}

bool optDirAverageNormals(KdTreeExtra *kdtree, const SurfaceParams & sp, const Vector3D & x, Vector3D * xnormal) {
	int i, surfel_len = kdtree->getNOfFoundNeighbours();
	float weight = 0;
	float dist_squared, totalweight = 0;
	float param = sp.gaussFactorSquared;
	Vector3D cummulative(0,0,0), v;

	for (i=0; i<surfel_len; i++) {
		int index = kdtree->getNeighbourPositionIndex(i);
		const Vector3D & pi = kdtree->getNeighbourPositionByIndex(index);
		const Vector3D & ni = kdtree->getNeighbourNormalByIndex(index);
		if (sp.useRadiusAsWeight) {
			param = kdtree->getNeighbourRadiusByIndex(index);
		}

		//<a,pi-x>^2
		v = pi - x;
		// Delta(x,pi) weight
		dist_squared = v.getSquaredLength();
		weight = exp(- dist_squared / param);
		totalweight += weight;
		cummulative += ni*weight;

		//since we are using exp fall off and the distances are ordered,
		//i + 1 weights will also have all zeros.
		if (weight == 0)
			break;
	}

	if (totalweight <= 0) {
		int index = kdtree->getNeighbourPositionIndex(0);
		*xnormal = kdtree->getNeighbourNormalByIndex(index);
	}
	else {
		(*xnormal) = cummulative;
		xnormal->normalize();
	}

	return true;
}

bool optDirMLS(KdTreeExtra *kdtree, const SurfaceParams & sp, const Vector3D & x, Vector3D * xnormal) {
	// matrix stuff:
	gsl_matrix *data_matrix;
	gsl_matrix *transposed_data_matrix;
	gsl_matrix *covariance_matrix;
	gsl_matrix *eigen_vectors;
	gsl_vector *eigen_values;
	gsl_eigen_symmv_workspace *workspace;
	// other
	float *weights;
	float totalweight;
	
	unsigned int i, knext = kdtree->getNOfFoundNeighbours();
	// matrix init:
	data_matrix = gsl_matrix_alloc(3,knext);
	transposed_data_matrix = gsl_matrix_alloc(knext, 3);
	covariance_matrix = gsl_matrix_alloc(3,3);
	eigen_vectors = gsl_matrix_alloc(3,3);
	eigen_values = gsl_vector_alloc(3);
	workspace = gsl_eigen_symmv_alloc(3);
		
	weights = new float[knext];
	_MCnew( (unsigned long)weights, sizeof(weights), __FILE__, __LINE__);

	totalweight = 0;
	Vector3D center(0.,0.,0.);
	
	if (sp.gaussFactorSquared== 0) {
		gsl_matrix_free(data_matrix);
		gsl_matrix_free(transposed_data_matrix);
		gsl_matrix_free(covariance_matrix);
		gsl_matrix_free(eigen_vectors);
		gsl_vector_free(eigen_values);
		gsl_eigen_symmv_free(workspace);

		_MCdelete((unsigned long)weights);
		delete[] weights;

		return false;
	}

	for (i=0; i<knext; i++) {
		Vector3D vec = kdtree->getNeighbourPosition(i);
		// get weights
		float current_distance = Vector3D::distance(vec, x);
		weights[i] = exp(-(current_distance*current_distance) / sp.gaussFactorSquared);
		totalweight += weights[i];
		// get data matrix entries
		gsl_matrix_set(data_matrix,0,i,vec[0]);
		gsl_matrix_set(data_matrix,1,i,vec[1]);
		gsl_matrix_set(data_matrix,2,i,vec[2]);
		// update center
		center += vec * weights[i];
	}

	if (totalweight == 0 || i<3) {
		gsl_matrix_free(data_matrix);
		gsl_matrix_free(transposed_data_matrix);
		gsl_matrix_free(covariance_matrix);
		gsl_matrix_free(eigen_vectors);
		gsl_vector_free(eigen_values);
		gsl_eigen_symmv_free(workspace);

		_MCdelete((unsigned long)weights);
		delete[] weights;
		
		return false;
	}

	center /= totalweight;

	// set up data matrix
	for (i=0; i<knext; i++) {
		gsl_matrix_set(data_matrix, 0, i, (gsl_matrix_get(data_matrix,0,i) - center[0])*weights[i]);
		gsl_matrix_set(data_matrix, 1, i, (gsl_matrix_get(data_matrix,1,i) - center[1])*weights[i]);
		gsl_matrix_set(data_matrix, 2, i, (gsl_matrix_get(data_matrix,2,i) - center[2])*weights[i]);
	}
	
	// create transposed data matrix
	gsl_matrix_transpose_memcpy(transposed_data_matrix, data_matrix);
	// create matrix views
	gsl_matrix_const_view data_matrix_view = gsl_matrix_const_submatrix(data_matrix, 0,0, 3, knext);
	gsl_matrix_const_view transposed_data_matrix_view = gsl_matrix_const_submatrix(transposed_data_matrix, 0,0, knext, 3);
	// get covariance matrix
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, &data_matrix_view.matrix, &transposed_data_matrix_view.matrix, 0.0, covariance_matrix);
	// calculate eigenvectors/eigenvalues of the covariance matrix
	gsl_eigen_symmv(covariance_matrix, eigen_values, eigen_vectors, workspace);
	gsl_eigen_symmv_sort(eigen_values, eigen_vectors, GSL_EIGEN_SORT_VAL_ASC);
	
//	_eigenvalues[0] = gsl_vector_get(eigen_values, 0);
//	_eigenvalues[1] = gsl_vector_get(eigen_values, 1);
//	_eigenvalues[2] = gsl_vector_get(eigen_values, 2);
	
	// set up plane equation
	(*xnormal)[0] = gsl_matrix_get(eigen_vectors,0,0);
	(*xnormal)[1] = gsl_matrix_get(eigen_vectors,1,0);
	(*xnormal)[2] = gsl_matrix_get(eigen_vectors,2,0);
	xnormal->normalize();

	gsl_matrix_free(data_matrix);
	gsl_matrix_free(transposed_data_matrix);
	gsl_matrix_free(covariance_matrix);
	gsl_matrix_free(eigen_vectors);
	gsl_vector_free(eigen_values);
	gsl_eigen_symmv_free(workspace);

	_MCdelete((unsigned long)weights);
	delete[] weights;

	return true;
}


