// Title:   KdTreeExtra.h
// Created: Thu Sep 25 15:15:19 2003
// Author: Yong Joo Kil
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
// Copyright (c) 2004, Institute for Data Analysis and Visualization, UC Davis.
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __KDTREE_EXTRA_H__
#define __KDTREE_EXTRA_H__

class SurfelInterface;

#include "../../../../Core/DataStructures/src/NeighbourHood.h"

//////////////////////////////////////////////////////////////
// class KdTreeExtra
//
// This class extends KdTree such that it contains not only point information, 
// but also normal, radius, etc..
//
//////////////////////////////////////////////////////////////

class KdTreeExtra : public KdTree {
public:
	/**
	 * Creates a k-d tree from the positions
	 *
	 * @param positions
	 *			point positions
	 * @param nOfPositions
	 *			number of points
	 * @param maxBucketSize
	 *			number of points per bucket
	 */
	inline KdTreeExtra(const Vector3D *positions, const unsigned int nOfPositions, const unsigned int maxBucketSize);
	inline KdTreeExtra(const Vector3D *positions, const Vector3D *normals, SurfelInterface **, const float * radius, const unsigned int nOfPositions, const unsigned int maxBucketSize);
	/**
	 * Destructor
	 */
	virtual ~KdTreeExtra();

	/** 
	 * get the normal of the i-th nearest neighbour
	 * i must be smaller than the number of found neighbours
	 *
	 * @param i
	 *			index of the nearest neighbour
	 * @return the position of the i-th nearest neighbour
	 */
	//These expects the input index to be given by getNeighbourPositionIndex(neighbourIndex)
	inline const Vector3D	&	getNeighbourPositionByIndex(const unsigned int index);
	inline const Vector3D	&	getNeighbourNormalByIndex(const unsigned int index);
	inline SurfelInterface *	getNeighbourSurfelByIndex(const unsigned int index);
	inline float				getNeighbourRadiusByIndex(const unsigned int index);

private:
	const Vector3D*				m_positions_public;   //remember that these are not allocated within this class, only referenced.
	const Vector3D*				m_normals;   //remember that these are not allocated within this class, only referenced.
	SurfelInterface**			m_surfels;   //remember that these are not allocated within this class, only referenced.
	const float *				m_radius;	//remember that these are not allocated within this class, only referenced.
};

inline KdTreeExtra::KdTreeExtra(const Vector3D *positions, const unsigned int nOfPositions, const unsigned int maxBucketSize)
:KdTree(positions, nOfPositions, maxBucketSize)
{
	m_normals				= NULL;
	m_surfels				= NULL;
	m_radius				= NULL;
	m_positions_public		= NULL;
}

inline KdTreeExtra::KdTreeExtra(const Vector3D *positions, const Vector3D *normals, SurfelInterface ** surfels, const float * radius, const unsigned int nOfPositions, const unsigned int maxBucketSize)
:KdTree(positions, nOfPositions, maxBucketSize)
{
	m_positions_public		= positions;
	m_normals				= normals;
	m_surfels				= surfels;
	m_radius				= radius;
}

inline KdTreeExtra::~KdTreeExtra() {
}

//These expects the input index to be given by getNeighbourPositionIndex(neighbourIndex)

inline const Vector3D & KdTreeExtra::getNeighbourPositionByIndex(const unsigned int index) {
	return m_positions_public[index];
}

inline const Vector3D & KdTreeExtra::getNeighbourNormalByIndex(const unsigned int index) {
	return m_normals[index];
}

inline SurfelInterface * KdTreeExtra::getNeighbourSurfelByIndex(const unsigned int index) {
	return m_surfels[index];
}

inline float KdTreeExtra::getNeighbourRadiusByIndex(const unsigned int index) {
	return m_radius[index];
}

/*
private member...
inline const Vector3D & KdTreeExtra::getNeighbourPositionByIndex(const unsigned int index) {
	return m_positions[index];
}
*/

#endif

