// Title:   PlaneFit.cpp
// Created: Thu Sep 25 15:15:54 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifdef _WIN32
	// turn off std::map warnings
	# pragma warning(push)
	# pragma warning(disable:4786) 
#endif


#include "PlaneFit.h"

//#include <qgl.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_eigen.h>

#include "Selector.h"

#include "NewOverload.h"

PlaneFit::PlaneFit() {
	_normal = Vector3D(0.,0.,0.);
	_isvalid = false;
}

void PlaneFit::doPlaneFit(const Vector3D& source, int knext, float param) {
	// matrix stuff:
	gsl_matrix *data_matrix;
	gsl_matrix *transposed_data_matrix;
	gsl_matrix *covariance_matrix;
	gsl_matrix *eigen_vectors;
	gsl_vector *eigen_values;
	gsl_eigen_symmv_workspace *workspace;
	// other
	Selector *selector;
	KdTree *kdtree;

	int i,counter;
	int index;
	float *weights;
	float totalweight;
	
	// matrix init:
	data_matrix = gsl_matrix_alloc(3,knext);
	transposed_data_matrix = gsl_matrix_alloc(knext, 3);
	covariance_matrix = gsl_matrix_alloc(3,3);
	eigen_vectors = gsl_matrix_alloc(3,3);
	eigen_values = gsl_vector_alloc(3);
	workspace = gsl_eigen_symmv_alloc(3);
	// other
	selector = Selector::getInstance();
	kdtree = selector->returnKdTree();

	if (!kdtree) {
		gsl_matrix_free(data_matrix);
		gsl_matrix_free(transposed_data_matrix);
		gsl_matrix_free(covariance_matrix);
		gsl_matrix_free(eigen_vectors);
		gsl_vector_free(eigen_values);
		gsl_eigen_symmv_free(workspace);

		failureHandler();
		return;
	}

	kdtree->setNOfNeighbours(knext*2);
	kdtree->queryPosition(source);
	
	weights = new float[knext];
	_MCnew( (unsigned long)weights, sizeof(weights), __FILE__, __LINE__);

	totalweight = 0;
	_center = Vector3D(0.,0.,0.);
	
	if (param == 0) {
		gsl_matrix_free(data_matrix);
		gsl_matrix_free(transposed_data_matrix);
		gsl_matrix_free(covariance_matrix);
		gsl_matrix_free(eigen_vectors);
		gsl_vector_free(eigen_values);
		gsl_eigen_symmv_free(workspace);

		_MCdelete((unsigned long)weights);
		delete[] weights;

		failureHandler();
		return;
	}
		
	for (i=0,counter=0; counter<knext; i++) {
		if (counter>=knext)
			qDebug("error error error in planefit");
		
		index = kdtree->getNeighbourPositionIndex(i);
		if (index>=0 && index<selector->getNOfSurfels() && selector->isValid(index)) {
			Vector3D vec = selector->getSurfelWorldPosition(index);
			// get weights
			float current_distance = Vector3D::distance(vec, source);
			weights[counter] = exp(-(current_distance*current_distance) / param);
			totalweight += weights[counter];
			// get data matrix entries
			gsl_matrix_set(data_matrix,0,counter,vec[0]);
			gsl_matrix_set(data_matrix,1,counter,vec[1]);
			gsl_matrix_set(data_matrix,2,counter,vec[2]);
			// update center
			_center += vec * weights[counter];
			counter++;
		}
	}

	if (totalweight == 0) {
		gsl_matrix_free(data_matrix);
		gsl_matrix_free(transposed_data_matrix);
		gsl_matrix_free(covariance_matrix);
		gsl_matrix_free(eigen_vectors);
		gsl_vector_free(eigen_values);
		gsl_eigen_symmv_free(workspace);

		_MCdelete((unsigned long)weights);
		delete[] weights;

		failureHandler();
		return;
	}

	_center /= totalweight;
	
	// set up data matrix
	for (i=0; i<counter; i++) {
		gsl_matrix_set(data_matrix, 0, i, (gsl_matrix_get(data_matrix,0,i) - _center[0])*weights[i]);
		gsl_matrix_set(data_matrix, 1, i, (gsl_matrix_get(data_matrix,1,i) - _center[1])*weights[i]);
		gsl_matrix_set(data_matrix, 2, i, (gsl_matrix_get(data_matrix,2,i) - _center[2])*weights[i]);
	}
	
	// create transposed data matrix
	gsl_matrix_transpose_memcpy(transposed_data_matrix, data_matrix);
	// create matrix views
	gsl_matrix_const_view data_matrix_view = gsl_matrix_const_submatrix(data_matrix, 0,0, 3, counter);
	gsl_matrix_const_view transposed_data_matrix_view = gsl_matrix_const_submatrix(transposed_data_matrix, 0,0, counter, 3);
	// get covariance matrix
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, &data_matrix_view.matrix, &transposed_data_matrix_view.matrix, 0.0, covariance_matrix);
	// calculate eigenvectors/eigenvalues of the covariance matrix
	gsl_eigen_symmv(covariance_matrix, eigen_values, eigen_vectors, workspace);
	gsl_eigen_symmv_sort(eigen_values, eigen_vectors, GSL_EIGEN_SORT_VAL_ASC);
	
	_eigenvalues[0] = gsl_vector_get(eigen_values, 0);
	_eigenvalues[1] = gsl_vector_get(eigen_values, 1);
	_eigenvalues[2] = gsl_vector_get(eigen_values, 2);
	
	// set up plane equation
	_normal[0] = gsl_matrix_get(eigen_vectors,0,0);
	_normal[1] = gsl_matrix_get(eigen_vectors,1,0);
	_normal[2] = gsl_matrix_get(eigen_vectors,2,0);
	_normal.normalize();

	_d = -1.*(_normal[0]*_center[0]+_normal[1]*_center[1]+_normal[2]*_center[2]);
	_isvalid = true;

	// clean up
	gsl_matrix_free(data_matrix);
	gsl_matrix_free(transposed_data_matrix);
	gsl_matrix_free(covariance_matrix);
	gsl_matrix_free(eigen_vectors);
	gsl_vector_free(eigen_values);
	gsl_eigen_symmv_free(workspace);

	_MCdelete((unsigned long)weights);
	delete[] weights;

}



void PlaneFit::doPlaneFitByKdTree(KdTree *kdtree, const Vector3D &source, int knext, float param) {
	// matrix stuff:
	gsl_matrix *data_matrix;
	gsl_matrix *transposed_data_matrix;
	gsl_matrix *covariance_matrix;
	gsl_matrix *eigen_vectors;
	gsl_vector *eigen_values;
	gsl_eigen_symmv_workspace *workspace;
	// other
	int i;
	float *weights;
	float totalweight;
	
	// matrix init:
	data_matrix = gsl_matrix_alloc(3,knext);
	transposed_data_matrix = gsl_matrix_alloc(knext, 3);
	covariance_matrix = gsl_matrix_alloc(3,3);
	eigen_vectors = gsl_matrix_alloc(3,3);
	eigen_values = gsl_vector_alloc(3);
	workspace = gsl_eigen_symmv_alloc(3);
		
	weights = new float[knext];
	_MCnew( (unsigned long)weights, sizeof(weights), __FILE__, __LINE__);

	totalweight = 0;
	_center = Vector3D(0.,0.,0.);
	
	if (param == 0) {
		gsl_matrix_free(data_matrix);
		gsl_matrix_free(transposed_data_matrix);
		gsl_matrix_free(covariance_matrix);
		gsl_matrix_free(eigen_vectors);
		gsl_vector_free(eigen_values);
		gsl_eigen_symmv_free(workspace);

		_MCdelete((unsigned long)weights);
		delete[] weights;

		failureHandler();
		return;
	}

	kdtree->setNOfNeighbours(knext+1);
	kdtree->queryPosition(source);
	if (kdtree->getNOfFoundNeighbours() < knext)
		knext = kdtree->getNOfFoundNeighbours();

	for (i=0; i<knext; i++) {
		Vector3D vec = kdtree->getNeighbourPosition(i);
		// get weights
		float current_distance = Vector3D::distance(vec, source);
		weights[i] = exp(-(current_distance*current_distance) / param);
		totalweight += weights[i];
		// get data matrix entries
		gsl_matrix_set(data_matrix,0,i,vec[0]);
		gsl_matrix_set(data_matrix,1,i,vec[1]);
		gsl_matrix_set(data_matrix,2,i,vec[2]);
		// update center
		_center += vec * weights[i];
	}

	if (totalweight == 0 || i<3) {
		gsl_matrix_free(data_matrix);
		gsl_matrix_free(transposed_data_matrix);
		gsl_matrix_free(covariance_matrix);
		gsl_matrix_free(eigen_vectors);
		gsl_vector_free(eigen_values);
		gsl_eigen_symmv_free(workspace);

		_MCdelete((unsigned long)weights);
		delete[] weights;
		
		failureHandler();
		return;
	}

	_center /= totalweight;

	// set up data matrix
	for (i=0; i<knext; i++) {
		gsl_matrix_set(data_matrix, 0, i, (gsl_matrix_get(data_matrix,0,i) - _center[0])*weights[i]);
		gsl_matrix_set(data_matrix, 1, i, (gsl_matrix_get(data_matrix,1,i) - _center[1])*weights[i]);
		gsl_matrix_set(data_matrix, 2, i, (gsl_matrix_get(data_matrix,2,i) - _center[2])*weights[i]);
	}
	
	// create transposed data matrix
	gsl_matrix_transpose_memcpy(transposed_data_matrix, data_matrix);
	// create matrix views
	gsl_matrix_const_view data_matrix_view = gsl_matrix_const_submatrix(data_matrix, 0,0, 3, knext);
	gsl_matrix_const_view transposed_data_matrix_view = gsl_matrix_const_submatrix(transposed_data_matrix, 0,0, knext, 3);
	// get covariance matrix
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, &data_matrix_view.matrix, &transposed_data_matrix_view.matrix, 0.0, covariance_matrix);
	// calculate eigenvectors/eigenvalues of the covariance matrix
	gsl_eigen_symmv(covariance_matrix, eigen_values, eigen_vectors, workspace);
	gsl_eigen_symmv_sort(eigen_values, eigen_vectors, GSL_EIGEN_SORT_VAL_ASC);
	
	_eigenvalues[0] = gsl_vector_get(eigen_values, 0);
	_eigenvalues[1] = gsl_vector_get(eigen_values, 1);
	_eigenvalues[2] = gsl_vector_get(eigen_values, 2);
	
	// set up plane equation
	_normal[0] = gsl_matrix_get(eigen_vectors,0,0);
	_normal[1] = gsl_matrix_get(eigen_vectors,1,0);
	_normal[2] = gsl_matrix_get(eigen_vectors,2,0);
	_normal.normalize();

	_d = -1.*(_normal[0]*_center[0]+_normal[1]*_center[1]+_normal[2]*_center[2]);
	_isvalid = true;

	gsl_matrix_free(data_matrix);
	gsl_matrix_free(transposed_data_matrix);
	gsl_matrix_free(covariance_matrix);
	gsl_matrix_free(eigen_vectors);
	gsl_vector_free(eigen_values);
	gsl_eigen_symmv_free(workspace);

	_MCdelete((unsigned long)weights);
	delete[] weights;
}

void PlaneFit::doPlaneFitAdaptiveWeight(const Vector3D& source, int knext) {
	Selector *selector;
	KdTree *kdtree;
	float param;
	
	selector = Selector::getInstance();
	kdtree = selector->returnKdTree();
	if (!kdtree) {
		failureHandler();
		return;
	}

	kdtree->setNOfNeighbours(knext);
	kdtree->queryPosition(source);
	
	// calc param adaptive ( - ln 0.05 =~ 3) 
	param = kdtree->getSquaredDistance(knext-1) / 3.0;

	doPlaneFit(source,knext,param);
}


/**
 * this planefit is for MLSScatterer
 * the newly added points are included in a query grid -> they have to be used too.
 * it includes all points from the query grid, that are closer to the source than the
 * farthest neighbour of source.
 */
void PlaneFit::doPlaneFitAdaptiveWeightIncludeQueryGrid(const Vector3D &source, int knext, QueryGrid<uInt> *qgrid) {
	Selector *selector;
	KdTree *kdtree;
	std::vector<Vector3D> fit_positions;
	MinPriorityQueue<Vector3D, float> mpq;
	Vector3D pos;
	float param;
	float maxdist;
	unsigned i;
	
	selector = Selector::getInstance();
	kdtree = selector->returnKdTree();
	if (!kdtree) {
		failureHandler();
		return;
	}
	
	// build pqueue
	kdtree->setNOfNeighbours(knext+1);
	kdtree->queryPosition(source);
	maxdist = sqrt(kdtree->getSquaredDistance(knext));
	
	qgrid->setNOfNeighbours(knext+1);
	qgrid->queryAdvisoryAverageQueryRadius(maxdist);
	qgrid->queryRange(source,maxdist);

	mpq.init();
	mpq.setSize(knext + qgrid->getNOfFoundNeighbours());
	// insert knext
	for (i=1; i<knext; i++) {
		pos = kdtree->getNeighbourPosition(i);
		mpq.insert(pos,Vector3D::distance(pos,source));
	}
	// insert sfls from qggrid
	for (i=0; i<qgrid->getNOfFoundNeighbours(); i++) {
		pos = qgrid->getNeighbourPosition(i);
		mpq.insert(pos, Vector3D::distance(pos,source));
	}

	// get first k from pqueue
	for (i=0; i<knext-1; i++) {
		fit_positions.push_back(mpq.getMinIndex());
		mpq.removeMin();
	}

	float distance_square = Vector3D::distance(fit_positions[knext-1],source) * Vector3D::distance(fit_positions[knext-1],source);
	param = distance_square / 3.0;

	doPlaneFitByVectorsStdVec(fit_positions, source,param);	
	fit_positions.clear();
}

void PlaneFit::doPlaneFitByVectorsStdVec(const std::vector<Vector3D> &positions, const Vector3D &source, float param) {
	// matrix stuff:
	gsl_matrix *data_matrix;
	gsl_matrix *transposed_data_matrix;
	gsl_matrix *covariance_matrix;
	
	gsl_matrix *eigen_vectors;
	gsl_vector *eigen_values;
	gsl_eigen_symmv_workspace *workspace;
	// other
	Vector3D vec;
	
	int i;
	int no_of_points;
	
	no_of_points = positions.size();
	if (no_of_points<=0) {
		failureHandler();
		return;
	}
	
	// matrix init:
	data_matrix = gsl_matrix_alloc(3,no_of_points);
	transposed_data_matrix = gsl_matrix_alloc(no_of_points, 3);	
	covariance_matrix = gsl_matrix_alloc(3,3);
	eigen_vectors = gsl_matrix_alloc(3,3);
	eigen_values = gsl_vector_alloc(3);
	workspace = gsl_eigen_symmv_alloc(3);
		
	_center = Vector3D(0.,0.,0.);
	
	for (i = 0; i<no_of_points; i++) {
		vec = positions[i];
		// get data matrix entries
		gsl_matrix_set(data_matrix,0,i,vec[0]);
		gsl_matrix_set(data_matrix,1,i,vec[1]);
		gsl_matrix_set(data_matrix,2,i,vec[2]);
		// update center
		_center += vec;
	}
	
	if (no_of_points == 0) {
		gsl_matrix_free(data_matrix);
		gsl_matrix_free(transposed_data_matrix);
		gsl_matrix_free(covariance_matrix);
		gsl_matrix_free(eigen_vectors);
		gsl_vector_free(eigen_values);
		gsl_eigen_symmv_free(workspace);

		failureHandler();
		return;
	}
	_center /= no_of_points;

	// set up data matrix
	for (i=0; i<no_of_points; i++) {
		gsl_matrix_set(data_matrix, 0, i, (gsl_matrix_get(data_matrix,0,i) - _center[0]));
		gsl_matrix_set(data_matrix, 1, i, (gsl_matrix_get(data_matrix,1,i) - _center[1]));
		gsl_matrix_set(data_matrix, 2, i, (gsl_matrix_get(data_matrix,2,i) - _center[2]));
	}
	
	// create transposed data matrix
	gsl_matrix_transpose_memcpy(transposed_data_matrix, data_matrix);

	// get covariance matrix
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, data_matrix, transposed_data_matrix, 0.0, covariance_matrix);
	// calculate eigenvectors/eigenvalues of the covariance matrix
	gsl_eigen_symmv(covariance_matrix, eigen_values, eigen_vectors, workspace);
	gsl_eigen_symmv_sort(eigen_values, eigen_vectors, GSL_EIGEN_SORT_VAL_ASC);

	_eigenvalues[0] = gsl_vector_get(eigen_values, 0);
	_eigenvalues[1] = gsl_vector_get(eigen_values, 1);
	_eigenvalues[2] = gsl_vector_get(eigen_values, 2);
	
	// set up plane equation
	_normal[0] = gsl_matrix_get(eigen_vectors,0,0);
	_normal[1] = gsl_matrix_get(eigen_vectors,1,0);
	_normal[2] = gsl_matrix_get(eigen_vectors,2,0);
	_normal.normalize();

	_d = -1.*(_normal[0]*_center[0]+_normal[1]*_center[1]+_normal[2]*_center[2]);
	_isvalid = true;

	gsl_matrix_free(data_matrix);
	gsl_matrix_free(transposed_data_matrix);
	gsl_matrix_free(covariance_matrix);
	gsl_matrix_free(eigen_vectors);
	gsl_vector_free(eigen_values);
	gsl_eigen_symmv_free(workspace);
}

/*
**	unweighted
*/

void gsl_handler (const char * reason, 
              const char * file, 
              int line, 
              int gsl_errno) {
	qDebug(reason);
	qDebug(file);
	qDebug("in line %d, errornumber %d",line,gsl_errno);
}


void PlaneFit::doPlaneFitByIndices(QVector<unsigned int> *indices, int start, int end) {
	// matrix stuff:
	gsl_matrix *data_matrix;
	gsl_matrix *transposed_data_matrix;
	gsl_matrix *covariance_matrix;
	gsl_matrix *eigen_vectors;
	gsl_vector *eigen_values;
	gsl_eigen_symmv_workspace *workspace;
	// other
	Selector *selector;
	Vector3D vec;
	int i, index;
	int no_of_points;
	int valid_point_count;

//#define MY_GSL_DEBUG

#ifdef MY_GSL_DEBUG
	gsl_error_handler_t *olderrorhandler;

	olderrorhandler = gsl_set_error_handler(gsl_handler);
#endif

	no_of_points = end - start;
	if (no_of_points<=0) {
		failureHandler();
		return;
	}
	
	// matrix init:
	data_matrix = gsl_matrix_alloc(3,no_of_points);
	transposed_data_matrix = gsl_matrix_alloc(no_of_points, 3);	
	covariance_matrix = gsl_matrix_alloc(3,3);
	eigen_vectors = gsl_matrix_alloc(3,3);
	eigen_values = gsl_vector_alloc(3);
	workspace = gsl_eigen_symmv_alloc(3);
	// other
	selector = Selector::getInstance();
	
	_center = Vector3D(0.,0.,0.);
	valid_point_count = 0;
	for (i = start, index = 0; i<end; i++, index++) {
		if (selector->isValid((*indices->at(i)))) {
			vec = selector->getSurfelWorldPosition(*indices->at(i));
			// get data matrix entries
			gsl_matrix_set(data_matrix,0,valid_point_count,vec[0]);
			gsl_matrix_set(data_matrix,1,valid_point_count,vec[1]);
			gsl_matrix_set(data_matrix,2,valid_point_count,vec[2]);
			// update center
			_center += vec;
			valid_point_count ++;
		}
	}
	
	if (valid_point_count == 0) {
		gsl_matrix_free(data_matrix);
		gsl_matrix_free(transposed_data_matrix);
		gsl_matrix_free(covariance_matrix);
		gsl_matrix_free(eigen_vectors);
		gsl_vector_free(eigen_values);
		gsl_eigen_symmv_free(workspace);

		failureHandler();
		return;
	}

	_center /= valid_point_count;

	// set up data matrix
	for (i=0; i<valid_point_count; i++) {
		gsl_matrix_set(data_matrix, 0, i, (gsl_matrix_get(data_matrix,0,i) - _center[0]));
		gsl_matrix_set(data_matrix, 1, i, (gsl_matrix_get(data_matrix,1,i) - _center[1]));
		gsl_matrix_set(data_matrix, 2, i, (gsl_matrix_get(data_matrix,2,i) - _center[2]));
	}
	
	// create transposed data matrix
	gsl_matrix_transpose_memcpy(transposed_data_matrix, data_matrix);
	
	// create data matrix views
	gsl_matrix_const_view data_matrix_view = gsl_matrix_const_submatrix(data_matrix, 0,0, 3,valid_point_count);
	gsl_matrix_const_view transposed_data_matrix_view = gsl_matrix_const_submatrix(transposed_data_matrix, 0,0, valid_point_count, 3);
	
	// get covariance matrix
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, &data_matrix_view.matrix, &transposed_data_matrix_view.matrix, 0.0, covariance_matrix);
	// calculate eigenvectors/eigenvalues of the covariance matrix
	gsl_eigen_symmv(covariance_matrix, eigen_values, eigen_vectors, workspace);
	gsl_eigen_symmv_sort(eigen_values, eigen_vectors, GSL_EIGEN_SORT_VAL_ASC);

	_eigenvalues[0] = gsl_vector_get(eigen_values, 0);
	_eigenvalues[1] = gsl_vector_get(eigen_values, 1);
	_eigenvalues[2] = gsl_vector_get(eigen_values, 2);
	
	// set up plane equation
	_normal[0] = gsl_matrix_get(eigen_vectors,0,0);
	_normal[1] = gsl_matrix_get(eigen_vectors,1,0);
	_normal[2] = gsl_matrix_get(eigen_vectors,2,0);
	_normal.normalize();

	_d = -1.*(_normal[0]*_center[0]+_normal[1]*_center[1]+_normal[2]*_center[2]);
	_isvalid = true;

#ifdef MY_GSL_DEBUG
	gsl_set_error_handler(olderrorhandler);
#endif
	gsl_matrix_free(data_matrix);
	gsl_matrix_free(transposed_data_matrix);
	gsl_matrix_free(covariance_matrix);
	gsl_matrix_free(eigen_vectors);
	gsl_vector_free(eigen_values);
	gsl_eigen_symmv_free(workspace);

}


void PlaneFit::doPlaneFitByIndicesStdVec(std::vector<unsigned int> *indices, int start, int end) {
	// matrix stuff:
	gsl_matrix *data_matrix;
	gsl_matrix *transposed_data_matrix;
	gsl_matrix *covariance_matrix;
	
	gsl_matrix *eigen_vectors;
	gsl_vector *eigen_values;
	gsl_eigen_symmv_workspace *workspace;
	// other
	Selector *selector;
	Vector3D vec;
	int i, index;
	int no_of_points;
	int valid_point_count;

	no_of_points = end - start;
	if (no_of_points<=0) {
		failureHandler();
		return;
	}
	
	// matrix init:
	data_matrix = gsl_matrix_alloc(3,no_of_points);
	transposed_data_matrix = gsl_matrix_alloc(no_of_points, 3);	
	covariance_matrix = gsl_matrix_alloc(3,3);
	eigen_vectors = gsl_matrix_alloc(3,3);
	eigen_values = gsl_vector_alloc(3);
	workspace = gsl_eigen_symmv_alloc(3);
	// other
	selector = Selector::getInstance();
	
	_center = Vector3D(0.,0.,0.);
	valid_point_count = 0;
	for (i = start, index = 0; i<end; i++, index++) {
		if (selector->isValid((*indices)[i])) {
			vec = selector->getSurfelWorldPosition((*indices)[i]);
			// get data matrix entries
			gsl_matrix_set(data_matrix,0,valid_point_count,vec[0]);
			gsl_matrix_set(data_matrix,1,valid_point_count,vec[1]);
			gsl_matrix_set(data_matrix,2,valid_point_count,vec[2]);
			// update center
			_center += vec;
			valid_point_count ++;
		}
	}

	if (valid_point_count < 3) { // not enough points to fit it!
		gsl_matrix_free(data_matrix);
		gsl_matrix_free(transposed_data_matrix);
		gsl_matrix_free(covariance_matrix);
		gsl_matrix_free(eigen_vectors);
		gsl_vector_free(eigen_values);
		gsl_eigen_symmv_free(workspace);

		failureHandler();
		return;
	}

	_center /= valid_point_count;

	// set up data matrix
	for (i=0; i<valid_point_count; i++) {
		gsl_matrix_set(data_matrix, 0, i, (gsl_matrix_get(data_matrix,0,i) - _center[0]));
		gsl_matrix_set(data_matrix, 1, i, (gsl_matrix_get(data_matrix,1,i) - _center[1]));
		gsl_matrix_set(data_matrix, 2, i, (gsl_matrix_get(data_matrix,2,i) - _center[2]));
	}
	
	// create transposed data matrix
	gsl_matrix_transpose_memcpy(transposed_data_matrix, data_matrix);

	// create matrix views
	gsl_matrix_const_view data_matrix_view = gsl_matrix_const_submatrix(data_matrix, 0,0, 3, valid_point_count);
	gsl_matrix_const_view transposed_data_matrix_view = gsl_matrix_const_submatrix(transposed_data_matrix, 0,0, valid_point_count, 3);

	// get covariance matrix
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, &data_matrix_view.matrix, &transposed_data_matrix_view.matrix, 0.0, covariance_matrix);
	// calculate eigenvectors/eigenvalues of the covariance matrix
	gsl_eigen_symmv(covariance_matrix, eigen_values, eigen_vectors, workspace);
	gsl_eigen_symmv_sort(eigen_values, eigen_vectors, GSL_EIGEN_SORT_VAL_ASC);

	_eigenvalues[0] = gsl_vector_get(eigen_values, 0);
	_eigenvalues[1] = gsl_vector_get(eigen_values, 1);
	_eigenvalues[2] = gsl_vector_get(eigen_values, 2);
	
	// set up plane equation
	_normal[0] = gsl_matrix_get(eigen_vectors,0,0);
	_normal[1] = gsl_matrix_get(eigen_vectors,1,0);
	_normal[2] = gsl_matrix_get(eigen_vectors,2,0);
	_normal.normalize();

	_d = -1.*(_normal[0]*_center[0]+_normal[1]*_center[1]+_normal[2]*_center[2]);
	_isvalid = true;

	gsl_matrix_free(data_matrix);
	gsl_matrix_free(transposed_data_matrix);
	gsl_matrix_free(covariance_matrix);
	gsl_matrix_free(eigen_vectors);
	gsl_vector_free(eigen_values);
	gsl_eigen_symmv_free(workspace);

}

void PlaneFit::getEigenValues(float eigenvals[3]) const {
	eigenvals[0] = _eigenvalues[0];
	eigenvals[1] = _eigenvalues[1];
	eigenvals[2] = _eigenvalues[2];
}

Vector3D PlaneFit::getNormal() const {
	return _normal;
}

Vector3D PlaneFit::getCenter() const {
	return _center;
}

float PlaneFit::getD() const  {
	return _d;
}

void PlaneFit::reset() {
	_normal = Vector3D(0.,0.,0.);
}

float PlaneFit::getDistance(Vector3D source) const {
	return Vector3D::dotProduct((source - _center),_normal);
}

bool PlaneFit::isValid() const {
	return _isvalid;
}

void PlaneFit::failureHandler() {
	_isvalid = false;
	_center = _normal = Vector3D(0,0,0);
	_d = 0;
	_eigenvalues[0] = _eigenvalues[1] = _eigenvalues[2] = 0;
}


// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
