// Title:   PlaneFit.h
// Created: Thu Sep 25 15:15:52 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __PLANEFIT_H
#define __PLANEFIT_H

#include <vector>
#include "../../../../DataTypes/src/Vector3D.h"
#include "../../../../Core/DataStructures/src/QueryGrid.h"
#include "../../../../Core/DataStructures/src/NeighbourHood.h"

#if QT_VERSION >= 0x030000
# include <qptrvector.h>
# ifndef QVector
#  define QVector QPtrVector
# endif
# ifndef QVectorIterator
#  define QVectorIterator QPtrVectorIterator
# endif
#else
# include <qvector.h>
#endif

/**
 * Plane fit class
 *	Calculates a plane of form:
 *		n[0]*x + n[1]*y + n[2]*z + d = 0;
 *
 * @author Simon Heinzle
 * @version 1.0
 */

typedef unsigned int* uIntPtr;
typedef unsigned int  uInt;

class PlaneFit {
	Vector3D _normal;
	Vector3D _center;
	float _d;
	float _eigenvalues[3];
	bool _isvalid;
	
	/**
	 * internal failure handler
	 */
	void failureHandler();
public:
	/**
	 * Constructor
	 */
	PlaneFit();
	/**
	 * Performs a plane fit based on the surfels contained by the Selector.
	 * Plane fit is weighted according to the distance to <code>const Vector3D& source</code>, 
	 * the <code>int knext</code> are contributing to this plane, weighted with exp^(- distance^2 / param^2) where <code>int param</code>
	 * defines this gauss parameter.
	 */
	void doPlaneFit(const Vector3D& source, int knext, float param);
	/**
	 * Performs a plane fit based on the surfels contained by the Selector.
	 * Weighting is done adaptive.
	 * Plane fit is weighted according to the distance to <code>const Vector3D& source</code>,
	 * the <code>int knext</code> are contributing to this plane
	 */
	void doPlaneFitAdaptiveWeight(const Vector3D& source, int knext);
	/**
	 * Performs a plane fit based on a query grid <code> QueryGrid<uInt> *qgrid </code>
	 * Plane fit is weighted according to the distance to <code>const Vector3D& source</code>,.
	 * The newly added points are included in a query grid -> they have to be used too.
	 * <code>int knext</code> defines the number of neigbours to be used.
	 */
	void doPlaneFitAdaptiveWeightIncludeQueryGrid(const Vector3D &source, int knext, QueryGrid<uInt> *qgrid);
	/**
	 * Performs a plane fit based on a stl vector <code>const std::vector<Vector3D> &positions</code>.
	 * Plane fit is weighted according to the distance to <code>const Vector3D& source</code>, 
	 * weighted with exp^(- distance^2 / param^2) where <code>int param</code>.
	 * All positions are used for plane fit.
	 */
	void doPlaneFitByVectorsStdVec(const std::vector<Vector3D> &positions, const Vector3D &source, float param);
	/**
	 * Performs a plane fit based on a kdtree.
	 * Plane fit is weighted according to the distance to <code>const Vector3D& source</code>, 
	 * the <code>int knext</code> are contributing to this plane, weighted with exp^(- distance^2 / param^2) where <code>int param</code>
	 * defines this gauss parameter.
	 */
	void doPlaneFitByKdTree(KdTree *kdtree, const Vector3D &source, int knext, float param);
	/** 
	 * Performs a unweighted plane fit based on given positions.
	 * Positions are given via <code>QVector<unsigned int> *indices</code>, which are indices into the surfels
	 * contained by the Selector.
	 * <code>int start</code> and <code>int end</code> indicate the start&end index for the <code>QVector<unsigned int> *indices</code>.
	 */
	void doPlaneFitByIndices(QVector<unsigned int> *indices, int start, int end);
	/** 
	 * Performs a unweighted plane fit based on given positions.
	 * Positions are given via <code>std::vector<unsigned int> *indices</code>, which are indices into the surfels
	 * contained by the Selector.
	 * <code>int start</code> and <code>int end</code> indicate the start&end index for the <code>std::vector<unsigned int> *indices</code>.
	 */
	void doPlaneFitByIndicesStdVec(std::vector<unsigned int> *indices, int start, int end);
	/**
	 * Returns the eigenvalues of the PCA in <code>float eigenvals[3]</code>
	 */
	void getEigenValues(float eigenvals[3]) const;
	/**
	 * @return the normal of the plane fit
	 */
	Vector3D getNormal() const;
	/**
	 * @return the center of the plane fit
	 */
	Vector3D getCenter() const;
	/**
	 * @return the d parameter of the plane
	 */
	float getD() const;
	/**
	 * Resets the plane fit
	 */
	void reset();
	/**
	 * @return the distance from <code>Vector3D source</code> to the plane
	 */
	float getDistance(Vector3D source) const;
	/**
	 * @return <code>bool</code> to indicate if planefit is valid.
	 */
	bool isValid() const;
};

#endif

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
