// Title:   PointNormalEnergy.cpp
// Created: Thu Sep 25 15:11:17 2003
// Authors: Yong Joo Kil (modified by)
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include "PointNormalEnergy.h"
#include "DefiningSurfaceTypes.h"
#include "Energies.h"
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_eigen.h>

#include "NewOverload.h"

#define ITMAX 100
#define ITMAX2 200		// here ITMAX is the maximum allowed number of iterations
#define ZEPS 1.0e-10	// EPS is a small number to rectify the special case of converging to exactly zero function value.
#define EPS 1.0e-10
#define TOL 2.0e-4f 		// Tolerance passed to dbrent.
#define MOV3(a,b,c, d,e,f) (a)=(d);(b)=(e);(c)=(f);
#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-20
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
static double maxarg1,maxarg2;
#define FMAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ? (maxarg1) : (maxarg2))

////////////////////////////////////////////
//
//	evaluateFunction(t)
//	where q = (sourcePoint + t*normal)
////////////////////////////////////////////
float PointNormalEnergy::evaluateFunction(float t) {
	Vector3D q = _sourcePoint + t*this->_normal;
	return _efunc(_kdtree, _sp, q, this->_normal);
}

float PointNormalEnergy::find_best_t(const Vector3D & sourcePoint, const Vector3D & normal) {
	float t = 0, val;
	this->_sourcePoint = sourcePoint;
	this->_normal = normal;
	float bound = _sp.brentSearchRange;
	float tol = _sp.brentTol;

	//Use minimizeBrent() if analytical derivative is known.  I personally have not seen much 
	//improvements by using the derivative.  Also, it's easy to make mistakes finding analytical solution,
	//esp. during experimentations.  However, the energy functions I use tends to look like 
	//a quadratic function (approximates distance squared), which should work nicely with derivative 
	//information...
	val = brent(-bound, 0.0f, bound, &PointNormalEnergy::evaluateFunction, tol, &t);
	return t;
}

Vector3D PointNormalEnergy::find_best_n(const Vector3D & source) {
	bool valid;
	Vector3D n;
	valid = _nfunc(_kdtree, _sp, source, &n);
	//do something if invalid
	return n;
}

/**
* "Numerical Recipes in C" - Routine
*
* Given a function f and its derivative function df, and given a bracketing triplet of abscissas ax,
* bx, cx [such that bx is between ax and cx, and f(bx) is less than both f(ax) and f(cx)],
* this routine isolates the minimum to a fractional precision of about tol using a modification of
* Brent's method that uses derivatives. The abscissa of the minimum is returned as xmin, and
* the minimum function value is returned as dbrent, the returned function value.
*/
//find minimum to be f(xmin)
float PointNormalEnergy::minimizeBrent(float ax, float bx, float cx, float (PointNormalEnergy::*f)(float), float (PointNormalEnergy::*df)(float), float tol, float *xmin) {

	// parts of the code from Numerical Recipes in C, 2nd Edition, pp 405ff

	int iter, ok1, ok2; 	// Will be used as flags for whether proposed steps are acceptable or not.
	float a, b, d, d1, d2, du, dv, dw, dx, e = 0.0;
	float fu, fv, fw, fx, olde, tol1, tol2, u, u1, u2, v, w, x, xm;

	a = (ax < cx ? ax : cx);
	b = (ax > cx ? ax : cx);
	x = w = v = bx;
	fw = fv = fx = (*this.*f)(x);
	dw = dv = dx = (*this.*df)(x);

	// All our housekeeping chores are doubled by the necessity of moving
	// derivative values around as well as function values.

	for (iter = 1; iter <= ITMAX; iter++) {
		xm = 0.5 * (a + b);
		tol1 = tol * fabs(x) + ZEPS;
		tol2 = 2.0 * tol1;

		if (fabs(x - xm) <= (tol2 - 0.5*(b - a))) {
			*xmin=x;
			return fx;
		}

		if (fabs(e) > tol1) {	// Initialize these d's to an out-of-bracket value.
			d1 = 2.0 * (b-a);
			d2 = d1;
			if (dw != dx) {
				d1 = (w - x)*dx/(dx - dw); // Secant method with one point.
			}
			if (dv != dx) {
				d2 = (v - x)*dx/(dx - dv); // And the other.
			}

			// Which of these two estimates of d shall we take? We will insist that they be within
			// the bracket, and on the side pointed to by the derivative at x:

			u1 = x + d1;
			u2 = x + d2;
			ok1 = (a - u1)*(u1 - b) > 0.0 && dx*d1 <= 0.0;
			ok2 = (a - u2)*(u2 - b) > 0.0 && dx*d2 <= 0.0;
			olde = e;	// Movement on the step before last.
			e = d;

			// Take only an acceptable d, and if both are acceptable, then take the smallest one.

			if (ok1 || ok2) {
				if (ok1 && ok2) {
					d = (fabs(d1) < fabs(d2) ? d1 : d2);
				}
				else if (ok1) {
					d = d1;
				}
				else {
					d = d2;
				}
				if (fabs(d) <= fabs(0.5 * olde)) {
					u = x + d;
					if (u-a < tol2 || b-u < tol2) {
						d = SIGN(tol1, xm - x);
					}
				}
				else {	//  Bisect, not golden section.
					d = 0.5 * (e = (dx >= 0.0 ? a - x : b - x));
					//Decide which segment by the sign of the derivative.
				}
			}
			else {
				d = 0.5 * (e = (dx >= 0.0 ? a - x : b - x));
			}
		}
		else {
			d = 0.5 * (e = (dx >= 0.0 ? a - x : b - x));
		}
		if (fabs(d) >= tol1) {
			u = x + d;
			fu = (*this.*f)(u);
		}
		else {
			u = x + SIGN(tol1,d);
			fu = (*this.*f)(u);

			// If the minimum step in the downhill direction takes us uphill, then we are done.
			if (fu > fx) {
				*xmin = x;
				return fx;
			}
		}
		du = (*this.*df)(u);	// Now all the housekeeping, sigh.

		if (fu <= fx) {
			if (u >= x) {
				a = x;
			}
			else {
				b = x;
			}
			MOV3(v,fv,dv,w,fw,dw)
			MOV3(w,fw,dw,x,fx,dx)
			MOV3(x,fx,dx,u,fu,du)
		}
		else {
			if (u < x) {
				a = u;
			}
			else {
				b = u;
			}
			if (fu <= fw || w == x) {
				MOV3(v,fv,dv,w,fw,dw)
				MOV3(w,fw,dw,u,fu,du)
			}
			else if (fu < fv || v == x || v == w) {
				MOV3(v,fv,dv,u,fu,du)
			}
		}
	}
	qDebug("Too many iterations in routine minimizeBrent");
	return 0.0; 	// Never get here.
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
