// Title:   PointNormalEnergy.h
// Created: Thu Sep 25 15:11:13 2003
// Authors: Yong Joo Kil (modified by)
// Authors: Richard Keiser, Mark Pauly, Sascha Scandella
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __POINTNORMALENERGY_H_
#define __POINTNORMALENERGY_H_

// Parts of the code from Numerical Recipes in C, 2nd Edition. (see Chapter 10.1 - 10.6)
// Note: Conjugate Gradient Methods not evaluated yet
//
// author Sascha Scandella (code only documented marginally)
// For details refer to "Computing and Rendering Point Set Surfaces".

#include "PlaneFit.h"
#include "../../../../DataTypes/src/Vector3D.h"
#include "../../../../Core/DataStructures/src/NeighbourHood.h"
#include <math.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_eigen.h>
#include "KdTreeExtra.h"
#include "DefiningSurfaceTypes.h"

/**
 * This is a moving least squares implementation
 */
class PointNormalEnergy {

public:

   /**
	* @param nbHood
	*			kd-Tree for neighbour queries
	* @param _h
	*			parameter h, previously used for gaussian spead
	*/
	void init(KdTreeExtra *kdtree, const Vector3D & r, const Vector3D & normal, const SurfaceParams & sp, 
		EnergyFunc ef, OptDirectionFunc nf) {
		this->_kdtree = kdtree;
		this->_sourcePoint = r;
		this->_sp = sp;
		this->_efunc = ef;
		this->_nfunc = nf;
	};

	PointNormalEnergy(KdTreeExtra *kdtree, const SurfaceParams & sp, EnergyFunc ef, OptDirectionFunc nf) {
		init(kdtree, Vector3D(0.0f, 0.0f, 0.0f), Vector3D(0.0f, 0.0f, 0.0f), sp, ef, nf);
	};

	~PointNormalEnergy() {};

	float evaluateFunction(float t);
	float evaluateDerivative(float t);


   /**
	* "Numerical Recipes in C" - Routine
	*
	* Given a function f and its derivative function df, and given a bracketing triplet of abscissas ax,
	* bx, cx [such that bx is between ax and cx, and f(bx) is less than both f(ax) and f(cx)],
	* this routine isolates the minimum to a fractional precision of about tol using a modification of
	* Brent's method that uses derivatives. The abscissa of the minimum is returned as xmin, and
	* the minimum function value is returned as dbrent, the returned function value.
	*/
	float minimizeBrent(float ax, float bx, float cx, float (PointNormalEnergy::*f)(float), float (PointNormalEnergy::*df)(float), float tol, float *xmin);
	float brent(float ax, float bx, float cx, float (PointNormalEnergy::*f)(float), float tol, float *xmin);
	float dbrent(float ax, float bx, float cx, float (PointNormalEnergy::*f)(float),float (PointNormalEnergy::*df)(float), float tol, float *xmin);

   /**
    * Method for finding the optimal direction, find_best_n, and the minimum along the optimal direction, find_best_t.
	* 
	*/
	Vector3D find_best_n(const Vector3D & sourcePoint);
	float find_best_t(const Vector3D & sourcePoint, const Vector3D & normal);

public:

	KdTreeExtra *_kdtree;

	Vector3D			_sourcePoint;
	Vector3D			_normal;
	SurfaceParams		_sp;
	EnergyFunc			_efunc;
	OptDirectionFunc	_nfunc;
};

#endif	// __POINTNORMALENERGY_H_
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
