// Title:   PointNormalSurface.cpp
// Created: Thu Sep 25 15:16:27 2003
// Authors: Yong Joo Kil (modified by)
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifdef _WIN32
	// turn off std::map warnings
	# pragma warning(disable:4786) 
#endif

#include "PointNormalSurface.h"
#include "../../../../DataTypes/src/MyDataTypes.h"
#include "../../../../Utilities/src/Matrix.h"
#include <gsl/gsl_multifit.h>
#include <gsl/gsl_errno.h>

#include "Energies.h"
#include "NewOverload.h"

PointNormalSurface::PointNormalSurface(const SurfaceParams & sp) {
	_kdtree = 0;
	_kdtree_positions = 0;
	_kdtree_normals = 0;
	_kdtree_surfels = 0;

	_sp = sp;
	switch(_sp.extremalSurfaceMethod) {
	case kExtremalSurfaceMahalanobis:
		setFunctions(energyMahalanobis, optDirAverageNormals);
		break;
	case kExtremalSurfaceMLS:
		setFunctions(energyMLS, optDirMLS);
		break;

	//Setup your own functions
	case kExtremalSurfaceX1:
		setFunctions(energyDummy, optDirDummy);
		break;
	default:
		setFunctions(energyDummy, optDirDummy);
		break;
	}
}

PointNormalSurface::~PointNormalSurface() {
	if (_kdtree) {
		_MCdelete((unsigned long)_kdtree);
		delete _kdtree;
		_kdtree = 0;
	}
	kdtree_cache_clean();
}

void PointNormalSurface::kdtree_cache_clean() {
	if (_kdtree_positions) {
		_MCdelete((unsigned long)_kdtree_positions);
		delete[] _kdtree_positions;
		_kdtree_positions = 0;
	}
	if (_kdtree_normals) {
		_MCdelete((unsigned long)_kdtree_normals);
		delete[] _kdtree_normals;
		_kdtree_normals = 0;
	}
	if (_kdtree_surfels) {
		_MCdelete((unsigned long)_kdtree_surfels);
		delete[] _kdtree_surfels;
		_kdtree_surfels = 0;
	}
}

void PointNormalSurface::setSource(const Vector3D &source) {
	_source = source;
}

void PointNormalSurface::doPlaneProjectionEnergyRealParticle() {
	PointNormalSurfacePlaneFit pf;
	float epsilon;
	int current_iteration;
	Vector3D new_source;

	current_iteration = 0;
	_plane_projected_position = _source;
	//setup kdtree
	_kdtree->setNOfNeighbours(_sp.neighbourhoodSize);

	do {
		_kdtree->queryRange(_plane_projected_position,FLT_MAX);
		pf.doPlaneFitByKdTreeEnergyRealParticle(_plane_projected_position, _kdtree, _sp, _ef, _nf);
		new_source = pf.getCenter();
		epsilon = Vector3D::distance(new_source, _plane_projected_position)/_plane_projected_position.getLength(); //why divide here?
		_plane_projected_position = new_source;
		current_iteration ++;
	} while ( (epsilon>_sp.brentTol) && (current_iteration < _sp.brentMaxIter) && (pf.getNormal().getLength() != 0.0f));
	
	if (!(current_iteration < _sp.brentMaxIter))
		while(0){};

	if (pf.getNormal().getLength() == 0)
		while(0){};

	_plane_fit = pf;
	return;
}

void PointNormalSurface::doPolynomialProjection(int degree) {
	// the vars for lsq fit
	gsl_matrix *X, *cov;
	gsl_vector *y,*c,*w;
	double chisquare;
	
	Vector3D u_direction;
	Vector3D v_direction;
	Vector3D normal;
	float z_coord;
	unsigned int i, knext = _kdtree->getNOfFoundNeighbours();
	unsigned int no_of_coeffs;

	// normal computation
	Vector3D tmp_normal;
	
	// check if valid planefit has occurred
	if (_plane_fit.getNormal().getLength() == 0.0f)
		return;

	normal = _plane_fit.getNormal();
	z_coord = (float) -1. / normal[2] * ( (normal[0] * 2 + normal[1] * 2 + _plane_fit.getD()) - (normal[0] + normal[1] + _plane_fit.getD()));
	u_direction = Vector3D(1.,1.,z_coord);
	v_direction = Vector3D(0.,0.,0.);
	
	u_direction.normalize();
	v_direction = Vector3D::crossProduct(u_direction, normal);
	v_direction.normalize();
	
	_kdtree->setNOfNeighbours(_sp.neighbourhoodSize+1);
	_kdtree->queryPosition(_plane_projected_position);

	// IMPORTANT! if point count < (degree+1)^2 we cannot fit!
	no_of_coeffs = (degree+1)*(degree+1);
	if (_kdtree->getNOfFoundNeighbours() < no_of_coeffs)
		return;

	// setting up matrices for lsq fit of a function of order degree:
	// f(x,y) = a x^degree y^degree +  ...
	X = gsl_matrix_alloc(_sp.neighbourhoodSize,no_of_coeffs);
	y = gsl_vector_alloc(_sp.neighbourhoodSize);
	w = gsl_vector_alloc(_sp.neighbourhoodSize);
	c = gsl_vector_alloc(no_of_coeffs);
	cov = gsl_matrix_alloc(no_of_coeffs,no_of_coeffs);
	
	for (i=0; i<knext; i++) {
		Vector3D point = _kdtree->getNeighbourPosition(i);
		float fi = Vector3D::dotProduct(point - _plane_projected_position, _plane_fit.getNormal());
		float ui = Vector3D::dotProduct(point - _plane_projected_position, u_direction);
		float vi = Vector3D::dotProduct(point - _plane_projected_position, v_direction);
		
		int c_it = no_of_coeffs - 1;
		float tmp_ui, tmp_vi;
		
		tmp_ui = 1.f;
		for (int u_it = 0; u_it <= degree; u_it++) {
			tmp_vi = 1.f;

			for (int v_it = 0; v_it <= degree; v_it++) {
				gsl_matrix_set(X, i, c_it, tmp_vi * tmp_ui);
				c_it --;
				tmp_vi *= vi;
			}

			tmp_ui *= ui;
		}
		
		gsl_vector_set (y, i, fi);
	
		float distance = Vector3D::distance(point, _plane_projected_position);
		float weight = (float) exp(-(distance*distance)/_sp.gaussFactorSquared);
		gsl_vector_set (w, i,weight);
	}

	// solving least squares equation
	gsl_multifit_linear_workspace  *workspace = gsl_multifit_linear_alloc (_sp.neighbourhoodSize, no_of_coeffs);
	gsl_multifit_wlinear(X, w, y, c, cov, &chisquare, workspace);
	
	
	_poly_projected_position = _plane_projected_position + _plane_fit.getNormal()* (float)gsl_vector_get(c,no_of_coeffs - 1);

	// getting the normal.
	float dds = (float)gsl_vector_get(c,no_of_coeffs - degree - 2);
	float ddt = (float)gsl_vector_get(c,no_of_coeffs - 2);

	Vector3D dxds;
	Vector3D dxdt;
	
	dxds = u_direction + Vector3D::crossProduct(u_direction,v_direction) * dds;
	dxdt = v_direction + Vector3D::crossProduct(u_direction,v_direction) * ddt;

	_normal = Vector3D::crossProduct(dxds,dxdt);
	_normal.normalize();

	gsl_matrix_free(X);
	gsl_matrix_free(cov);
	gsl_vector_free(y);
	gsl_vector_free(w);
	gsl_vector_free(c);
	gsl_multifit_linear_free(workspace);
}

//made into private function
//Note: this function assumes that the memory is allocated for the parameters before this function is called.
void PointNormalSurface::setPositionsNormals(const Vector3D *positions, const Vector3D *normals, SurfelInterface ** surfels, const float * radius, int no_of_positions) {
	if (_kdtree) {
		_MCdelete((unsigned int)_kdtree);
		delete _kdtree;
	}

	_kdtree = new KdTreeExtra(positions, normals, surfels, radius, no_of_positions, 10);
	_MCnew( (unsigned long)_kdtree, sizeof(_kdtree), __FILE__, __LINE__);
	_kdtree->setNOfNeighbours(_sp.neighbourhoodSize);
}

void PointNormalSurface::setPositionsNormals(const std::vector<Vector3D> &positions, const std::vector<Vector3D> &normals,
	const std::vector<SurfelInterface*> &surfels, const std::vector<float> &radius) {

	kdtree_cache_clean();

	_kdtree_positions = new Vector3D[positions.size()];
	_MCnew( (unsigned long)_kdtree_positions, sizeof(_kdtree_positions), __FILE__, __LINE__);
	_kdtree_normals = new Vector3D[normals.size()];
	_MCnew( (unsigned long)_kdtree_normals, sizeof(_kdtree_normals), __FILE__, __LINE__);
	_kdtree_surfels = new SurfelInterface*[surfels.size()];
	_MCnew( (unsigned long)_kdtree_surfels, sizeof(_kdtree_surfels), __FILE__, __LINE__);
	_kdtree_radius = new float[radius.size()];
	_MCnew( (unsigned long)_kdtree_radius, sizeof(_kdtree_radius), __FILE__, __LINE__);

	for (unsigned int i=0; i<positions.size(); i++) {
		_kdtree_positions[i] = positions[i];
		_kdtree_normals[i] = normals[i];
		_kdtree_surfels[i] = surfels[i];
		if (_sp.useRadiusAsWeight) {
			_kdtree_radius[i] = radius[i] * radius[i] * _sp.radiusScale * _sp.radiusScale;
		}
		else {
			_kdtree_radius[i] = _sp.gaussFactorSquared;
		}
	}
	setPositionsNormals(_kdtree_positions, _kdtree_normals, _kdtree_surfels, _kdtree_radius, positions.size());
}

Vector3D PointNormalSurface::getPlanePosition() const {
	return _plane_projected_position;
}

Vector3D PointNormalSurface::getPlaneNormal() const {
	return _plane_fit.getNormal();
}

Vector3D PointNormalSurface::getPolyPosition() const {
	return _poly_projected_position;
}

Vector3D PointNormalSurface::getPolyNormal() const {
	return _normal;
}

bool PointNormalSurface::isValid() const {
	return _plane_fit.isValid();
}


// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
