// Title:   PointNormalSurface.h
// Created: Thu Sep 25 15:16:26 2003
// Authors: Yong Joo Kil (modified by)
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifndef __POINTNORMAL_SURFACE_H
#define __POINTNORMAL_SURFACE_H

#ifdef _WIN32
	// turn off std::map warnings
	# pragma warning(push)
	# pragma warning(disable:4786) 
#endif

/**
 * YJK Point Normal Surface Projector
 *
 *
 * @author Simon Heinzle, Yong Joo Kil (modified by)
 * @version 1.0
 */

#include "PointNormalSurfacePlaneFit.h"

#include "../../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../../Core/Interfaces/src/SurfelInterface.h"
#include "KdTreeExtra.h"
#include "DefiningSurfaceTypes.h"

#include <vector>
#if QT_VERSION >= 0x030000
# include <qptrvector.h>
# ifndef QVector
#  define QVector QPtrVector
# endif
# ifndef QVectorIterator
#  define QVectorIterator QPtrVectorIterator
# endif
#else
# include <qvector.h>
#endif

class PointNormalSurface {
	Vector3D _source;
	// planefit for normal
	PointNormalSurfacePlaneFit _plane_fit;
	// projected position onto plane
	Vector3D _plane_projected_position;
	// projected position onto mls polynom
	Vector3D _poly_projected_position;
	Vector3D _normal;
	// parameters
	SurfaceParams _sp;
	EnergyFunc _ef;
	OptDirectionFunc _nf;

	// knext query tree
	KdTreeExtra *_kdtree;

	//actual memory is allocated at this level, not at the kdtree level.
	Vector3D *_kdtree_positions;
	Vector3D *_kdtree_normals;
	SurfelInterface **_kdtree_surfels;
	float *_kdtree_radius;

	void kdtree_cache_clean();

	/**
	 * Sets the positions which are used by the projection.
	 */
	void setPositionsNormals(const Vector3D *positions, const Vector3D *normals, SurfelInterface **, const float * radius, int no_of_positions);
	/**
	 * Sets the energy function and optimal direction functions.  This can be used to manually change 
	 * the functions.  However, this call will be made within the constructor, given the apporpriate SurfaceParams.
	 */
	inline void setFunctions(EnergyFunc ef, OptDirectionFunc nf) {
		_ef = ef;
		_nf = nf;
	}
public:
	/**
	 * Constructor
	 *
	 * Parameter <code>float min_error = 1.e-3f</code> defines the minimum error of the plane projection,
	 * <code>int max_iterations = 10</code> defines the maximum number of iterations for plane projection
	 * if <code>min_error </code>is not reached.
	 */
	PointNormalSurface(const SurfaceParams & sp);
	/**
	 * Deconstructor
	 */
	~PointNormalSurface();
	/**
	 * Sets the number of neighbours to be used for projection.
	 */
	inline void setParams(const SurfaceParams & sp) {
		_sp = sp;
	};
	inline SurfaceParams getParams() {
		return _sp;
	};
	/**
	 * Sets the <code>const Vector3D &source</code> as point to be projected.
	 */
	void setSource(const Vector3D &source);
	/**
	 * Sets the positions which are used by the projection.
	 */
	void setPositionsNormals(const std::vector<Vector3D> &positions, const std::vector<Vector3D> &normals, 
		const std::vector<SurfelInterface *> & surfels, const std::vector<float> & radius );
	/**
	 * Performs the plane projection.
	 */
	void doPlaneProjectionEnergyRealParticle();

	/**
	 * Performs the polynomial projection, works only if plane projection is done earlier.
	 * <code> int degree = 3 </code> defines the degree of the fitted polynom, must be > 0.
	 */
	void doPolynomialProjection(int degree = 3);
	/**
	 * Perform full (plane & polynomial) projection. 
	 * <code> int degree = 3 </code> defines the degree of the fitted polynom, must be > 0.
	 */
	/**
	 * @return the fitted plane projected position
	 */
	Vector3D getPlanePosition() const;
	/**
	 * @return the fitted plane projected normal
	 */
	Vector3D getPlaneNormal() const;
	/**
	 * @return the fitted polynomial projected position
	 */
	Vector3D getPolyPosition() const;
	/**
	 * @return the fitted polynomial projected normal
	 */
	Vector3D getPolyNormal() const;
	/**
	 * @return a <code>bool</code> to indicate wheter this projection is valid or not.
	 */
	bool isValid() const;

};

#ifdef _WIN32
	# pragma warning(pop)
#endif

#endif

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
