// Title:   ScanTools.cpp
// Created: Thu Sep 25 15:15:46 2003
// Authors: Simon Heinzle, Tim Weyrich
// Adoptation by: Yong Joo Kil
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifdef _WIN32
	// turn off std::map warnings
	//# pragma warning(push)
	# pragma warning(disable:4786) // identifier was truncated to 25 char...
#endif

#include <vector>
#include <qtoolbutton.h>
#include <qtoolbar.h>
#include <qcursor.h>
#include <qevent.h>
#include <qpixmap.h>
#include <qwhatsthis.h>
#include <qwidget.h>
#include <qpainter.h>
#include <qcolor.h>
#include <qlineedit.h>
#include <string.h>
#include <qlistbox.h>
#include <qtabwidget.h>
#include <qpushbutton.h>
#include <qcheckbox.h>
#include <qslider.h>
#include <qlabel.h>
#include <qradiobutton.h>
#include <qcombobox.h>

#include "../../../../Core/RendererManager/src/RendererManager.h"
#include "../../../../Core/Interfaces/src/ToolInterface.h"
#include "../UI/ScanToolsWidget.h"

#include "ScanTools.h"
#include "ScanToolsGlobals.h"

#ifndef _WIN32
#include <stdlib.h>
#define _gcvt gcvt
#endif

// ***************
// * tool includes
#include "Selector.h"
#include "DefiningSurfaces.h"
#include "Help.h"

#include "NewOverload.h"

// ******************
// 'What's this' text
// ******************

static const char *whatsThisText = "<img source=\"template\"> "
"Click this button to activate the Defining Surface Tool"
"<ul><li>Defining Surface Tool is designed to define various surface methods</li>";

// *****************************************************
// Help Text
// *****************************************************

QString selectionString = "Here you can set selections for projecting objects on to the surface.  This is also used to define selections as described in ScanTools plugin.";
QString mlsString = "Here you define which object(s) define the surface."\
	"Bigger gauss factor leads to more smoothing. The bigger the gauss factor, the more points influence the smoothing,"\
	"so for very big gauss factors i recommend increasing the k next neighbour count.\n"\
	"Degree of polynom defines the degree of the resulting fitted polynom (must be greater than 0)."\
	"Not using polynomail is equivalent to a plane fit / polynom degree 0.\n";

// ************
// Box selector
//  this global definition is due to callback passing
// ************

BoxSelector box;

void paint_box_callback() {
	box.paintBox();
}

// **********
// Tool class
// **********

ScanTools::ScanTools() {
	toolCursor = new QCursor (CrossCursor);
	_MCnew( (unsigned long)toolCursor, sizeof(toolCursor), __FILE__, __LINE__);

	scantoolsDialog = ScanToolsDialogImpl::getInstance();
	
	undoRedoStack = new ActionManager(4);
	_MCnew( (unsigned long)undoRedoStack, sizeof(undoRedoStack), __FILE__, __LINE__);

	box.reset();
	showGL();
	controlPressed = false;
	shiftPressed = false;
	altPressed = false;

#ifdef _DEBUG
	scantoolsDialog->memleak->setEnabled(true);
#endif
	// pass keys from widget
	this->connect (scantoolsDialog, SIGNAL (passKeyPressEvent(QKeyEvent*)),
				   this, SLOT (handleKeyPressEventFromDialog(QKeyEvent*)));
	this->connect (scantoolsDialog, SIGNAL (passKeyReleaseEvent(QKeyEvent*)),
				   this, SLOT (handleKeyReleaseEventFromDialog(QKeyEvent*)));
}

ScanTools::~ScanTools() {
	_MCdelete((unsigned int)toolCursor);
	delete toolCursor;
	ScanToolsDialogImpl::destroyInstance();
	hideGL();
	MemChecker::getInstance()->DumpUnfreed();
}

QToolButton *ScanTools::addButtonToToolBar (QToolBar *toolBar) {
	QPixmap markIcon ("Resources/DEF_MarkIcon.png"),
			mlsIcon("Resources/DEF_MLSIcon.png");

	// create icon sets
	markSet.setPixmap(markIcon, QIconSet::Automatic, QIconSet::Normal);
	mlsSet.setPixmap(mlsIcon, QIconSet::Automatic, QIconSet::Normal);

	// add the button to the toolbar
	toolButton = new QToolButton (markSet, "Scan Cleaning Tool Set", "Scan Cleaning Tool Set for processing 3d Scans",
	                              0, 0, toolBar, "Scan Cleaning Tools");
	_MCnew( (unsigned long)toolButton, sizeof(toolButton), __FILE__, __LINE__);
	toolButton->setToggleButton(true);
	
	QWhatsThis::add (toolButton, whatsThisText);
	QMimeSourceFactory::defaultFactory()->setPixmap ("scan cleaning tools", markSet.pixmap(QIconSet::Automatic, QIconSet::Normal));

	markID = scantoolsMenu.insertItem(markSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	mlsID = scantoolsMenu.insertItem(mlsSet.pixmap(QIconSet::Automatic, QIconSet::Normal), this, SLOT(handleToolSelected(int)));
	
	currentTool = markID;

	toolButton->setEnabled(true);
	
	isToolSetSelected = false;
	
	return toolButton;
}

QToolButton *ScanTools::getToolButton() {
	return toolButton;
}

QCursor *ScanTools::getToolCursor() const {
	return toolCursor;
}

QWidget *ScanTools::getToolConfigurationWidget() const {
	//return scantoolsConfigurationDialog;
//	return scantoolsBoxColorDialog;
	return NULL;
}

bool ScanTools::isAvailable() const {
	return this->checkToolAvailability();
}

void ScanTools::handleKeyPressEvent (QKeyEvent *ke) {
	if (ke->key() == Key_Control)
		controlPressed = true;
	if (ke->key() == Key_Shift)
		shiftPressed = true;
	if (ke->key() == Key_Alt)
		altPressed = true;
	
	if (controlPressed) {
		if (ke->key() == Key_Z)
			undoRedoStack->undo();
		else if (ke->key() == Key_Y)
			undoRedoStack->redo();
	} 

	ke->ignore();
}

void ScanTools::handleKeyReleaseEvent (QKeyEvent *ke) {
	if (ke->key() == Key_Control)
		controlPressed = false;
	if (ke->key() == Key_Shift)
		shiftPressed = false;
	if (ke->key() == Key_Alt)
		altPressed = false;


	ke->ignore();
}

void ScanTools::handleMousePressEvent (QMouseEvent *me) {
	if (!((me->state()&ShiftButton) || (me->state()&ControlButton) || (me->state()&AltButton))) {
		
		if ((currentTool == markID))
			box.setSelectionIndex(getCurrentSelectionIndex());
		else { 
			Selector::getInstance()->deselectIndex(3);
			box.setSelectionIndex(3);
		}
	}
	
	box.handleMousePress(me->x(), me->y(), stateToInt(me->stateAfter()));
	RendererManager::getInstance()->getCurrentRenderer()->renderFrame();
}

void ScanTools::handleMouseReleaseEvent (QMouseEvent *me) {
	box.handleMouseRelease(me->x(), me->y(), stateToInt(me->state()));
	
	if ( (me->state()&ShiftButton) || (me->state()&ControlButton) || (me->state()&AltButton) )
		return;
	if (! Scene::getInstance()->getActiveObject())
		return;
	if (scantoolsDialog->PickingButton->isChecked()) 
		return;

	if (currentTool == mlsID) {
		/*
		DefiningSurfaces mls;
		bool isnumber1, isnumber2, isnumber3, isnumber4;
		mls.setSelection(3);
		mls.setParameters( scantoolsDialog->MLSnEdit->text().toFloat(&isnumber1),
			scantoolsDialog->MLSgaussEdit->text().toFloat(&isnumber2),
			scantoolsDialog->MLSalphaEdit->text().toFloat(&isnumber3),
			scantoolsDialog->MLSdoplaneBox->isChecked(),
			scantoolsDialog->MLSdegreeEdit->text().toFloat(&isnumber4));
		if (isnumber1 && isnumber2 && isnumber3 && isnumber4)
			undoRedoStack->addUndoAction(mls.executeUndo());
		Selector::getInstance()->deselectIndex(3);
		*/
	}
	RendererManager::getInstance()->getCurrentRenderer()->renderFrame();
}


void ScanTools::handleMouseMoveEvent (QMouseEvent *me) {
	
	box.handleMouseMovementMidpoint(me->x(), me->y(), stateToInt(me->stateAfter()));
	RendererManager::getInstance()->getCurrentRenderer()->renderFrame();
}

void ScanTools::selectTool() {
	
	// set menu
	toolButton->setPopup(&scantoolsMenu);
	// update the dialog
	updateWindow();
	// switch to preview renderer
	RendererManager::getInstance()->setPreviewRendererEnabled(true,true);
	// show box
	showGL();
	// show dialog
	scantoolsDialog->show();
	scantoolsDialog->raise();
	// turn mouse tracking on
	RendererManager::getInstance()->getCurrentRenderer()->getRendererWidget()->setMouseTracking (true);
	// establish connections
	connectSelectionOptions();
	connectOther();

	// update object list
	updateObjects();
	//YJK: bug fix?
	handleObjectListSelectionChanged();

	// set scantoolsdialog apropriate
	currentTool = markID;
	toolButton->setIconSet(markSet);
	// enable / disable tabs
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab,true);
	scantoolsDialog->TabWidget->setTabEnabled(scantoolsDialog->tab_2,true);

	scantoolsDialog->MLSButton->setEnabled(true);
	// set apropriate page
	scantoolsDialog->TabWidget->setCurrentPage(1);
}

void ScanTools::deselectTool() {
	// turn of menu
	toolButton->setPopup(0);
	// turn mouse tracking off
	RendererManager::getInstance()->getCurrentRenderer()->getRendererWidget()->setMouseTracking (false);
	// hide dialog
	scantoolsDialog->hide();
	// hide box
	hideGL();
	// break connections
	disconnectSelectionOptions();
	disconnectOther();
}


// ************
// public slots
// ************

void ScanTools::handleKeyPressEventFromDialog(QKeyEvent* ke) {
	handleKeyPressEvent(ke);
}

void ScanTools::handleKeyReleaseEventFromDialog(QKeyEvent* ke) {
	handleKeyReleaseEvent(ke);
}
	

void ScanTools::handleRendererWidgetRepainted() {
	// the RenderWidget has repainted its content, so here this tool
	// can draw its content with a QPainter, if necessary
}

void ScanTools::handleToolSelected(int id) {
	currentTool = id;
	if (id == markID) {
		toolButton->setIconSet(markSet);
		scantoolsDialog->TabWidget->setCurrentPage(0);
	} else if (id == mlsID) {
		toolButton->setIconSet(mlsSet);
		scantoolsDialog->TabWidget->setCurrentPage(1);
	}
}

void ScanTools::handleObjectAdded() {
	updateObjects();
}

void ScanTools::handleObjectRemoved() {
	updateObjects();
}

void ScanTools::handleShapeBoxClicked(int id) {
	switch (id) {
		case 0:
			box.setShape(SHAPE_BOX);
			break;
		case 1:
			box.setShape(SHAPE_SPHERE);
			break;
	}
}

void ScanTools::handleSelectAllButtonReleased() {
	Selector::getInstance()->selectAll(getCurrentSelectionIndex());
}

void ScanTools::handleDeselectButtonReleased() {
	Selector::getInstance()->deselectIndex(getCurrentSelectionIndex());
}

void ScanTools::handleObjectListSelectionChanged() {
	Selector::getInstance()->removeAllObjects();
	
	for (unsigned int i=0; i<objects.count(); i++) {
		if (scantoolsDialog->ObjectList->isSelected(i)) 
			Selector::getInstance()->addObjectToSelection(objects[i]);
		if (scantoolsDialog->SurfaceObjectList->isSelected(i)) 
			Selector::getInstance()->addObjectToSurfaceDef(objects[i]);
	}
}

void ScanTools::handleMouseBoxClicked(int id) {
	switch(id) {
		case 0:
			box.setMouseMode(MOUSE_ALIGN);
			break;
		case 1:
			box.setMouseMode(MOUSE_FREEMOVE);
			break;
		case 2:
			box.setMouseMode(MOUSE_PICKING);
			break;
	}
}

void ScanTools::handleMLSButtonReleased() {
	DefiningSurfaces mls;
	bool isnum1, isnum2, isnum3 = true, isnum4, 
		isnum5, isnum6, isnum7, isnum8, isnum9;
	float alpha = 1.0f;

	mls.setSelection(getCurrentSelectionIndex());
	mls.setParameters( scantoolsDialog->MLSnEdit->text().toFloat(&isnum1),
		scantoolsDialog->MLSgaussEdit->text().toFloat(&isnum2),
		alpha,
		!scantoolsDialog->MLSdopolyBox->isChecked(),
		scantoolsDialog->MLSdegreeEdit->text().toFloat(&isnum4),

		//new params
		scantoolsDialog->ExtremalSurfaceMethodCombo->currentItem(),

		scantoolsDialog->WeightbyRadiusBox->isChecked(),
		scantoolsDialog->RadiusScaleEdt->text().toFloat(&isnum5),
		
		scantoolsDialog->MahalanobisScaleEdt->text().toFloat(&isnum6),

		scantoolsDialog->BrentMaxIterEdt->text().toFloat(&isnum7),
		scantoolsDialog->BrentTolEdt->text().toFloat(&isnum8),
		scantoolsDialog->BrentSearchRangeEdt->text().toFloat(&isnum9),
		scantoolsDialog->ProjectSelectionOnlyBox->isChecked()
		);
	if (isnum1 && isnum2 && isnum3 && isnum4 && isnum5 && isnum6 && isnum7 && isnum8 && isnum9)
		undoRedoStack->addUndoAction(mls.executeUndo());
}

void ScanTools::handleTabWidgetSelected(const QString& str) {
	if (str == "Remove Outliers") {
	}
}

void ScanTools::handleMemLeak() {
	MemChecker::getInstance()->DumpUnfreed();
}

// ***************
// private methods
// ***************


void ScanTools::updateObjects() {
	Scene *scene;
	Object *obj;
	unsigned int count = 0;

	scantoolsDialog->ObjectList->clear();
	scantoolsDialog->SurfaceObjectList->clear();
	Selector::getInstance()->removeAllObjects();
	
	scene = Scene::getInstance();
	
	objects.clear();
	objects.resize(scene->getNofObjects());
	
	count = 0;
	for (obj = scene->getFirstObject(); obj !=0; obj = scene->getNextObject()) {
		objects.insert(count, obj);
		scantoolsDialog->ObjectList->insertItem(obj->getName(), count);
		scantoolsDialog->SurfaceObjectList->insertItem(obj->getName(), count);

		//By default the first object is the surface surfels, while the rest are considered
		//points to be projected.
		if (count == 0) {
			scantoolsDialog->ObjectList->setSelected(count,false);
			scantoolsDialog->SurfaceObjectList->setSelected(count,true);		
		}
		else {
			scantoolsDialog->ObjectList->setSelected(count,true);
			scantoolsDialog->SurfaceObjectList->setSelected(count,false);		
		}
		count++;
	}
}

bool ScanTools::checkToolAvailability() const {
	
	// TODO: check here any conditions, e.g.
	//       - if an interactive main renderer is present
	//       - if an Object is in the Scene
	//       - etc.
	return true;

}

void ScanTools::updateWindow() {

}


void ScanTools::showGL(void) {
	RendererManager::getInstance()->getCurrentRenderer()->registerPostRenderCB(paint_box_callback);
	RendererManager::getInstance()->getCurrentRenderer()->renderFrame();
}

void ScanTools::hideGL(void) {
	RendererManager::getInstance()->getCurrentRenderer()->unregisterPostRenderCB(paint_box_callback);
	RendererManager::getInstance()->getCurrentRenderer()->renderFrame();
}


int ScanTools::stateToInt(ButtonState state) {
	int return_int = 0;
	
	if (state&LeftButton) 
		return_int = LEFTBUTTON;
	else if (state&MidButton)
		return_int = MIDBUTTON;
	else if (state&RightButton)
		return_int = RIGHTBUTTON;

	if (state&ControlButton)
		return_int |= CTRLBUTTON;
	if (state&AltButton)
		return_int |= ALTBUTTON;
	if (state&ShiftButton)
		return_int |= SHIFTBUTTON;
	
	return return_int;
}

void ScanTools::connectSelectionOptions() {
	// objects added / removed
	Scene *scene = Scene::getInstance();
	this->connect (scene, SIGNAL (objectAdded()),
		           this, SLOT (handleObjectAdded()));
	this->connect (scene, SIGNAL (objectRemoved()),
		           this, SLOT (handleObjectRemoved()));
	
	// ShapeBox clicked
	this->connect (scantoolsDialog->ShapeBox, SIGNAL (clicked(int)),
		           this, SLOT (handleShapeBoxClicked(int)));
	// SelectAllButton released
	this->connect (scantoolsDialog->SelectAllButton, SIGNAL (released()),
		           this, SLOT (handleSelectAllButtonReleased()));
	// DeselectButton released
	this->connect (scantoolsDialog->DeselectButton, SIGNAL (released()),
		           this, SLOT (handleDeselectButtonReleased()));
	// ObjectList selection changed
	this->connect (scantoolsDialog->ObjectList, SIGNAL (selectionChanged()),
		           this, SLOT (handleObjectListSelectionChanged()));
	this->connect (scantoolsDialog->SurfaceObjectList, SIGNAL (selectionChanged()),
		           this, SLOT (handleObjectListSelectionChanged()));
	// MouseBox clicked
	this->connect (scantoolsDialog->MouseBox, SIGNAL (clicked(int)),
		           this, SLOT (handleMouseBoxClicked(int)));

	this->connect (scantoolsDialog->memleak, SIGNAL(released()),
		this, SLOT(handleMemLeak()));

}

void ScanTools::disconnectSelectionOptions() {
	// objects added / removed
	Scene *scene = Scene::getInstance();
	this->disconnect (scene, SIGNAL (objectAdded()),
		              this, SLOT (handleObjectAdded()));
	this->disconnect (scene, SIGNAL (objectRemoved()),
		              this, SLOT (handleObjectRemoved()));

	// ShapeBox clicked
	this->disconnect (scantoolsDialog->ShapeBox, SIGNAL (clicked(int)),
		           this, SLOT (handleShapeBoxClicked(int)));
	// SelectAllButton released
	this->disconnect (scantoolsDialog->SelectAllButton, SIGNAL (released()),
		           this, SLOT (handleSelectAllButtonReleased()));
	// DeselectButton released
	this->disconnect (scantoolsDialog->DeselectButton, SIGNAL (released()),
		           this, SLOT (handleDeselectButtonReleased()));
	// ObjectList selection changed
	this->disconnect (scantoolsDialog->ObjectList, SIGNAL (selectionChanged()),
		           this, SLOT (handleObjectListSelectionChanged()));
	this->disconnect (scantoolsDialog->SurfaceObjectList, SIGNAL (selectionChanged()),
		           this, SLOT (handleObjectListSelectionChanged()));
	// MouseBox clicked
	this->disconnect (scantoolsDialog->MouseBox, SIGNAL (clicked(int)),
		           this, SLOT (handleMouseBoxClicked(int)));
}

void ScanTools::connectOther() {
	
	this->connect (scantoolsDialog->MLSButton, SIGNAL (released()),
		           this, SLOT (handleMLSButtonReleased()));
	this->connect (scantoolsDialog->TabWidget, SIGNAL (selected(const QString&)),
				   this, SLOT (handleTabWidgetSelected(const QString&)));

	// connect help
	this->connect (scantoolsDialog->HelpButtonSelection, SIGNAL (released()),
				   this, SLOT (handleHelpSelection()));
	this->connect (scantoolsDialog->HelpButtonMLS, SIGNAL (released()),
				   this, SLOT (handleHelpMLS()));
}

void ScanTools::disconnectOther() {
	
	this->disconnect (scantoolsDialog->MLSButton, SIGNAL (released()),
		           this, SLOT (handleMLSButtonReleased()));
	this->disconnect (scantoolsDialog->TabWidget, SIGNAL (selected(const QString&)),
				   this, SLOT (handleTabWidgetSelected(const QString&)));

	// disconnect help
	this->disconnect (scantoolsDialog->HelpButtonSelection, SIGNAL (released()),
				   this, SLOT (handleHelpSelection()));
	this->disconnect (scantoolsDialog->HelpButtonMLS, SIGNAL (released()),
				   this, SLOT (handleHelpMLS()));
}


int ScanTools::getCurrentSelectionIndex() {
	return scantoolsDialog->SelectionBox->id(scantoolsDialog->SelectionBox->selected());
}


/*

	// connect help
	this->connect (scantoolsDialog->HelpButtonSelection, SIGNAL (released()),
				   this, SLOT ());
	this->connect (scantoolsDialog->HelpButtonDeletor, SIGNAL (released()),
				   this, SLOT (handleHelpDeletor(deletorString)));
	this->connect (scantoolsDialog->HelpButtonNoise, SIGNAL (released()),
				   this, SLOT (handleHelpNoise(noiseString)));
	this->connect (scantoolsDialog->HelpButtonMLS, SIGNAL (released()),
				   this, SLOT (handleHelpMLS(mlsString)));
	this->connect (scantoolsDialog->HelpButtonScatterer, SIGNAL (released()),
				   this, SLOT (handleHelpScatterer(scattererString)));
	this->connect (scantoolsDialog->HelpButtonLPS, SIGNAL (released()),
				   this, SLOT (handleHelpLPS(lpsString)));
	this->connect (scantoolsDialog->HelpButtonOutliers, SIGNAL (released()),
				   this, SLOT (handleHelpOutliers(outlierString)));
	this->connect (scantoolsDialog->HelpButtonNormals, SIGNAL (released()),
				   this, SLOT (handleHelpNormals(normalsString)));
	this->connect (scantoolsDialog->HelpButtonResize, SIGNAL (released()),
				   this, SLOT (handleHelpResize(resizeString)));
	this->connect (scantoolsDialog->HelpButtonRecolor, SIGNAL (released()),
				   this, SLOT (handleHelpRecolor(recolorString)));
	
  */


void ScanTools::handleHelpSelection () {
	handleHelp(selectionString);
}

void ScanTools::handleHelpMLS () {
	handleHelp(mlsString);
}

void ScanTools::handleHelp(const QString& helpstring) {
	HelpDialogImpl *help = HelpDialogImpl::getInstance();
	help->showHelp(helpstring);
}

// ***************
// private slots
// ***************

void ScanTools::handleToolButtonDialog(void) {
}

// ****************
// access functions
// ****************

ToolInterface *createTool() {
	return new ScanTools();
}

const char *getToolName() {
	return "ScanTools";
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
