// Title:   Selector.cpp
// Created: Thu Sep 25 15:15:21 2003
// Authors: Simon Heinzle, Tim Weyrich
//
// Copyright (c) 2001, 2002, 2003, Computer Graphics Lab, ETH Zurich
//
// This file is part of the Pointshop3D system.
// See http://www.pointshop3d.com/ for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#ifdef _WIN32
	// turn off std::map warnings
	# pragma warning(push)
	# pragma warning(disable:4786) 
#endif

#include "Selector.h"

#include <math.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <algorithm>
#include <qapplication.h>
#if QT_VERSION >= 0x030000
# include <qptrlist.h>
# ifndef QList
#  define QList QPtrList
# endif
# ifndef QListIterator
#  define QListIterator QPtrListIterator
# endif
#else
# include <qlist.h>
#endif

#include "UndoRedoSpecialisations.h"

#define REBUILD_BORDER 100

#include "NewOverload.h"

Selector *_instance = 0;

//kept it as vector<> instead of map<> since object list probably won't be higher than 10. 
inline bool IsObjInObjList(Object * obj, const std::vector<Object *> & surfaceObjectList) {
	for (unsigned int i = 0; i < surfaceObjectList.size(); ++i) {
		if (obj == surfaceObjectList[i])
			return true;
	}
	return false;
}

Selector::Selector() {
	_index_bounds.clear();
	_valid_surfels_count = 0;
	_all_surfels_count = 0;
	_transformed_positions = 0;
	_transformed_normals =0;
	_valid_surfels = 0;
	_kdtree = 0;
	_selected_objects.clear();
	mSurfaceObjectList.clear();

	_render_manager = RendererManager::getInstance();
	_selection_tool = new SelectionTool();
	_MCnew( (unsigned long)_selection_tool, sizeof(_selection_tool), __FILE__, __LINE__);

	_object_selection_changed = false;
	
	buildTree();
}

Selector::~Selector() {
	if (_transformed_positions) {
		_MCdelete((unsigned long)_transformed_positions);
		delete[] _transformed_positions;
	}
	if (_transformed_normals) {
		_MCdelete((unsigned long)_transformed_normals);
		delete[] _transformed_normals;
	}
	if (_valid_surfels) {
		_MCdelete((unsigned long)_valid_surfels);
		delete[] _valid_surfels;
	}
	if (_kdtree) {
		_MCdelete((unsigned long)_kdtree);
		delete _kdtree;
	}
	if (_selection_tool) {
		_MCdelete((unsigned long)_selection_tool);
		delete _selection_tool;
	}
}

QVector<SurfelInterface>* Selector::getSphere(Vector3D center, float radius) {
	float max_squared_distance; // the squared distance we want to query
	unsigned int position;
	int found_count;
	Scene *scene;
	QVector<SurfelInterface> *tmp_vector;

	if (_all_surfels_count == 0 || ((float)_valid_surfels_count / (float)_all_surfels_count) < 0.5) {
		// if we're below some surfel count, don't rebuild tree
		if (_all_surfels_count > REBUILD_BORDER)
			buildTree();
	}

	if (_valid_surfels_count == 0)		// there aren't any surfels left
			return 0;

	
	// compute max squared distance
	max_squared_distance = radius;
	max_squared_distance *= max_squared_distance;

	// set the query range: 
	// NOTE: NOfNeighbours must be set to no. of surfels in order to get all neighbours.
	_kdtree->setNOfNeighbours(_all_surfels_count);
	_kdtree->queryRange(center,max_squared_distance);
	tmp_vector = new QVector<SurfelInterface>;
	_MCnew( (unsigned long)tmp_vector, sizeof(tmp_vector), __FILE__, __LINE__);

	tmp_vector->resize(_kdtree->getNOfFoundNeighbours());
	
	// query & return found surfels!
	scene = Scene::getInstance();
	found_count = 0;
	for (unsigned int i = 0; i<_kdtree->getNOfFoundNeighbours(); i++ ) {
		position = _kdtree->getNeighbourPositionIndex(i);
		if (position >= 0 && position<_all_surfels_count && _valid_surfels[position]) {
			tmp_vector->insert(found_count, _all_surfels[position]);
			found_count++;
		}
	}
	
	return tmp_vector;
}

std::vector<int>* Selector::getSphereIndicesStdVec(Vector3D center, float radius) {
	float max_squared_distance; // the squared distance we want to query
	unsigned int position;
	int found_count;
	Scene *scene;
	std::vector<int> *tmp_vector;

	if (_all_surfels_count == 0 || ((float)_valid_surfels_count / (float)_all_surfels_count) < 0.5) {
		// if we're below some surfel count, don't rebuild tree
		if (_all_surfels_count > REBUILD_BORDER)
			buildTree();
	}

	if (_valid_surfels_count == 0)		// there aren't any surfels left
			return 0;

	
	// compute max squared distance
	max_squared_distance = radius;
	max_squared_distance *= max_squared_distance;

	// set the query range: 
	// NOTE: NOfNeighbours must be set to no. of surfels in order to get all neighbours.
	_kdtree->setNOfNeighbours(_all_surfels_count);
	_kdtree->queryRange(center,max_squared_distance);
	tmp_vector = new std::vector<int>;
	_MCnew( (unsigned long)tmp_vector, sizeof(tmp_vector), __FILE__, __LINE__);

	tmp_vector->clear();
	
	// query & return found surfels!
	scene = Scene::getInstance();
	found_count = 0;
	for (unsigned int i = 0; i<_kdtree->getNOfFoundNeighbours(); i++ ) {
		position = _kdtree->getNeighbourPositionIndex(i);
		if (position >= 0 && position < _all_surfels_count && _valid_surfels[position]) {
			tmp_vector->push_back(position);
			found_count++;
		}
	}
	
	return tmp_vector;
}

void Selector::invalidateSurfel(unsigned int index) {
	if (index <=_all_surfels_count && _valid_surfels[index]) {
		_valid_surfels[index] = false;
		_valid_surfels_count--;
	}
}

void Selector::invalidateSurfels(SurfelInterface::Flag invalidate_flag) {
	for (unsigned int i=0; i<_all_surfels_count; i++) {
		if (
			(_valid_surfels[i])&&(_all_surfels[i]->isFlagOn(invalidate_flag)))
			invalidateSurfel(i);
	}
}

void Selector::deleteSurfels(std::vector<unsigned int>& indices) {
	std::vector<unsigned int>::iterator it;
	std::map<unsigned int,Object*>::iterator bound_it;
	unsigned int next_index_bound;
	Object *obj;
	SurfelCollection *coll;
	QList<SurfelInterface> sfls_to_delete;

	// sort indices
	std::sort(indices.begin(), indices.end());
	// get iterator to sorted boundaries
	bound_it = _index_bounds.begin();
	// adjust first obj
	next_index_bound = (*bound_it).first;
	obj = (*bound_it).second;
	coll = obj->getSurfelCollection();
	sfls_to_delete.clear();
	for (it = indices.begin(); it != indices.end(); it++) {
		if (next_index_bound < *it) {
			// delete surfels
			coll->removeSurfels(&sfls_to_delete, false);
			sfls_to_delete.clear();
			// get next boundaries and objec
			bound_it++;
			next_index_bound = (*bound_it).first;
			obj = (*bound_it).second;
			coll = obj->getSurfelCollection();
		}
		if (_valid_surfels[*it]) {
			sfls_to_delete.append(_all_surfels[*it]);
			_valid_surfels[*it] = false;
		}
	}
	if (sfls_to_delete.count())
		coll->removeSurfels(&sfls_to_delete,false);

	Scene::getInstance()->emitObjectModifiedSignal();
	Scene::getInstance()->emitRenderFrameSignal();
}

UndoAction* Selector::deleteSurfelsUndo(std::vector<unsigned int>& indices) {
	std::vector<SurfelInterface*> surfels;
	std::map<unsigned int, Object*> obj_bounds;
	std::map<unsigned int, Object*>::const_iterator bound_it;
	unsigned int index;
	unsigned int next_index_bound;
	Object *obj;
	SurfelCollection *coll;
	QList<SurfelInterface> sfls_to_delete;
	std::vector<unsigned int>::iterator it;
	DeleteUndo *undo;	
	
	// sort indices
	std::sort(indices.begin(), indices.end());
	
	// get iterator to sorted boundaries
	bound_it = _index_bounds.begin();
	
	next_index_bound = (*bound_it).first;
	obj = (*bound_it).second;
	index = 0;
	for (it = indices.begin(); it != indices.end(); it++) {
		if (next_index_bound < *it) {
			obj_bounds[index] = (*bound_it).second;
			bound_it++;
			next_index_bound = (*bound_it).first;
		}

		if (_valid_surfels[*it]) {
			surfels.push_back(_all_surfels[*it]);
			index++;
		}
	}
	obj_bounds[index] = (*bound_it).second;

	undo = new DeleteUndo(surfels,obj_bounds);
	_MCnew( (unsigned long)undo, sizeof(undo), __FILE__, __LINE__);
	
	// delete 'em 

	// get iterator to sorted boundaries
	bound_it = _index_bounds.begin();
	// adjust first obj
	next_index_bound = (*bound_it).first;
	obj = (*bound_it).second;
	coll = obj->getSurfelCollection();
	sfls_to_delete.clear();
	for (it = indices.begin(); it != indices.end(); it++) {
		if (next_index_bound < *it) {
			// delete surfels
			coll->removeSurfels(&sfls_to_delete, false);
			sfls_to_delete.clear();
			// get next boundaries and objec
			bound_it++;
			next_index_bound = (*bound_it).first;
			obj = (*bound_it).second;
			coll = obj->getSurfelCollection();
		}
		if (_valid_surfels[*it]) {
			sfls_to_delete.append(_all_surfels[*it]);
			_valid_surfels[*it] = false;
		}
	}
	if (sfls_to_delete.count())
		coll->removeSurfels(&sfls_to_delete,false);

	Scene::getInstance()->emitObjectModifiedSignal();
	Scene::getInstance()->emitRenderFrameSignal();

	return undo;

}

void Selector::deleteSurfelFromObject(SurfelInterface *surfel, Object *object, bool emitsignals) {
	SurfelCollection *coll;
	coll = object->getSurfelCollection();
	coll->removeSurfel(surfel, emitsignals);
}

SurfelInterface* Selector::insertSurfelToObject(SurfelInterface *surfel, Object *object, bool emitsignals) {
	SurfelCollection *coll;
	coll = object->getSurfelCollection();
	return coll->addSurfel(surfel,emitsignals);
}

bool Selector::isValid(unsigned int index) { 
	if (index >= _all_surfels_count)
		return false;

	return _valid_surfels[index]; 
}

Selector* Selector::getInstance(void) {
	if (_instance == 0) {
		_instance = new Selector;
		_MCnew( (unsigned long)_instance, sizeof(_instance), __FILE__, __LINE__);
	}
	return _instance;
}

void Selector::deleteInstance(void) {
	if (_instance) {
		_MCdelete((unsigned long)_instance);
		delete _instance;
	}
}

void Selector::buildTree(void) {
	Scene *scene;
	SurfelCollection *collection;
	SurfelInterface *surfel;
	Object *current_object;
	
	tTransformationPair ttp;
	MyDataTypes::TransformationMatrix16f trafo_matrix;
	MyDataTypes::TransformationMatrix16f invtrafo_matrix;
	std::map<Object*, tTransformationPair,MyObjectComparator>::iterator iter;
	Vector3D transformed_vector;
	
	// delete old arrays
	if (_kdtree) {
		_MCdelete((unsigned long)_kdtree);
		delete _kdtree;
		_kdtree = 0;
	}
	if (_transformed_positions) {
		_MCdelete((unsigned long)_transformed_positions);
		delete[] _transformed_positions;
		_transformed_positions = 0;
	}
	if (_valid_surfels) {
		_MCdelete((unsigned long)_valid_surfels);
		delete[] _valid_surfels;
		_valid_surfels = 0;
	}
	
	_index_bounds.clear();
	_all_surfels.clear();

	_all_surfels_count  = 0;
	_valid_surfels_count = 0;

	// get all surfels
	scene = Scene::getInstance();
	if (!scene)
		return;
		
	// get total nr. of surfels & build the index tree
	_all_surfels_count = 0;
//	for (current_object = getFirstSelectedObject(); current_object!=0; current_object = getNextSelectedObject()) {
//	YJK: collect all objects, regardless of what is selected; probably will break something else, but will work for 
//	my purpose.
	for (current_object = scene->getFirstObject(); current_object!=0; current_object = scene->getNextObject()) {
		_all_surfels_count += current_object->getSurfelCollection()->getNofSurfels();
		_index_bounds[_all_surfels_count-1] = current_object;
	}

	if (_all_surfels_count == 0)
		return;
	_valid_surfels_count  = _all_surfels_count;
	// set up containers
	_transformed_positions = new Vector3D[_all_surfels_count]; 
	_MCnew( (unsigned long)_transformed_positions, sizeof(_transformed_positions), __FILE__, __LINE__);

	_transformed_normals = new Vector3D[_all_surfels_count]; 
	_MCnew( (unsigned long)_transformed_normals, sizeof(_transformed_normals), __FILE__, __LINE__);
	
	_valid_surfels = new bool[_all_surfels_count];
	_MCnew( (unsigned long)_valid_surfels, sizeof(_valid_surfels), __FILE__, __LINE__);
	
	_all_surfels.resize(_all_surfels_count);
		
	// get all surfels & their positions
	int index = 0;
	
//	for (current_object = getFirstSelectedObject(); current_object!=0; current_object = getNextSelectedObject()) {
	for (current_object = scene->getFirstObject(); current_object!=0; current_object = scene->getNextObject()) {
		collection = current_object->getSurfelCollection();
		if (!collection)
			continue;
		current_object->getTransformationMatrix(trafo_matrix);
		// update matrices for objects.

	
//		iter = _selected_objects.find(current_object);
//		if (iter == _selected_objects.end()) 
//			continue;
		
		if (!MtrInverse4x4f(trafo_matrix, invtrafo_matrix))
			qDebug("Singular Transformation Matrix!!!");
		
		memcpy(ttp.transformation, trafo_matrix, 16*sizeof(float));
		memcpy(ttp.inverse_transformation, invtrafo_matrix, 16*sizeof(float));
		_selected_objects[current_object] = ttp;
		
		for (surfel = collection->getFirstSurfel(); surfel!=0; surfel = collection->getNextSurfel()) {
			MtrMultVector3DHom(trafo_matrix,surfel->getPosition(),&transformed_vector);
			_transformed_positions[index] = transformed_vector;

			MtrMultVector3DHom(trafo_matrix,surfel->getNormal(),&transformed_vector);
			_transformed_normals[index] = transformed_vector;

			_all_surfels.insert(index, surfel);
			_valid_surfels[index] = true;
			index ++;
		}
		// now check if index tree built correctly
		if ( (*_index_bounds.lower_bound(index - 1)).second != current_object) 
			qDebug("XXX, index tree not ok. %d is found upper index",(*_index_bounds.lower_bound(index - 1)).first);
	}

	// create kd-tree
	_kdtree = new KdTree(_transformed_positions,_all_surfels_count, 10);
	_MCnew( (unsigned long)_kdtree, sizeof(_kdtree), __FILE__, __LINE__);
	
}

int Selector::selectSphere(Vector3D center, float radius, int selection_index, bool reset_selection) {
	QApplication::setOverrideCursor(Qt::waitCursor);

	float max_squared_distance; // the squared distance we want to query
	unsigned int position;
	int found_count;
	Scene *scene;
	SurfelInterface::Flag flag_to_be_set;
	
	std::vector<SurfelInterface*> sfls;
	std::vector<int> flags;
	
	
	if(_object_selection_changed) {
		_object_selection_changed = false;
		buildTree();
	}

	if (reset_selection) {
		_selection_tool->setActiveSelection(selection_index);
		_selection_tool->clearSelection(false);
	}

	if (_all_surfels_count == 0 || ((float)_valid_surfels_count / (float)_all_surfels_count) < 0.5) {
		// if we're below some surfel count, don't rebuild tree
		if (_all_surfels_count > REBUILD_BORDER)
			buildTree();
	}

	if (_valid_surfels_count == 0) {		// there aren't any surfels left
		QApplication::restoreOverrideCursor();	
		return 0;
	}

	// compute max squared distance
	max_squared_distance = radius;
	max_squared_distance *= max_squared_distance;

	// set the query range: 
	// NOTE: NOfNeighbours must be set to no. of surfels in order to get all neighbours.
	_kdtree->setNOfNeighbours(_all_surfels_count);
	_kdtree->queryRange(center,max_squared_distance);
	
	// query & mark found surfels!
	scene = Scene::getInstance();
	found_count = 0;
	flag_to_be_set = intToFlag(selection_index);
	for (unsigned int i = 0; i<_kdtree->getNOfFoundNeighbours(); i++ ) {
		position = _kdtree->getNeighbourPositionIndex(i);
		if (position >= 0 && position < _all_surfels_count && _valid_surfels[position]) {
			scene->setSurfelFlagOn(_all_surfels[position], flag_to_be_set, true, false);
			scene->setSurfelFlagOn(_all_surfels[position], SurfelInterface::EMPHASISE, true, false);
			found_count++;
		}
	}
	
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	
	QApplication::restoreOverrideCursor();
	return found_count;
}

int Selector::selectBox(BoxSelector *boxselector,int selection_index, bool reset_selection) {
	QApplication::setOverrideCursor(Qt::waitCursor);

	gsl_matrix *box_transformation_matrix;
	gsl_matrix *inv_box_transformation_matrix;
	gsl_permutation *transformation_permutation; // for lu decomposition
	int signum; // for lu decomposition
	gsl_matrix *data_matrix;
	gsl_matrix *transformed_data_matrix;
	Scene *scene;
	float radius;
	unsigned int j;
	float x,y,z;
	int counter;
	SurfelInterface::Flag flag_to_be_set;
	std::vector<int> *selection_indices;

	std::vector<SurfelInterface*> sfls;
	std::vector<int> flags;

	
	//_last_action = 0;

	if(_object_selection_changed) {
		_object_selection_changed = false;
		buildTree();
	}

	if (reset_selection) {
		_selection_tool->setActiveSelection(selection_index);
		_selection_tool->clearSelection(false);
	}

	if (_all_surfels_count == 0 || ((float)_valid_surfels_count / (float)_all_surfels_count) < 0.5) {
		// if we're below some surfel count, don't rebuild tree
		if (_all_surfels_count > REBUILD_BORDER)
			buildTree();
	}

	if (_valid_surfels_count == 0) {		// there aren't any surfels left
		QApplication::restoreOverrideCursor();	
		return 0;
	}

	box_transformation_matrix = boxselector->getTransformationMatrixGSL();
	
	// calc inverse of box_transformation_matrix
	
	// init permutation vector for lu - decomposition
	transformation_permutation = gsl_permutation_alloc(4);
	gsl_permutation_init(transformation_permutation);
	// perform lu - decompositon
	gsl_linalg_LU_decomp (box_transformation_matrix,transformation_permutation, &signum);
	// calculate inverse transformation matrix (with lu - version of transf. matrix)
	inv_box_transformation_matrix = gsl_matrix_alloc(4,4);
	gsl_linalg_LU_invert (box_transformation_matrix, transformation_permutation, inv_box_transformation_matrix);

	// determine all surfels within range of diagonal-length
	radius = boxselector->getEdgeLength()[0];
	radius = boxselector->getEdgeLength()[1] > radius ? boxselector->getEdgeLength()[1] : radius;
	radius = boxselector->getEdgeLength()[2] > radius ? boxselector->getEdgeLength()[2] : radius;

	// use 1.5 instead of 2.f to get all possible inliers
	radius /= 1.5f;
	radius = (float) sqrt(2.f)*radius;

	//sphere_selection = getSphere(boxselector->getCenter(),radius);
	selection_indices = getSphereIndicesStdVec(boxselector->getCenter(), radius);
	
	if (selection_indices->size() == 0) {
		QApplication::restoreOverrideCursor();
		_MCdelete((unsigned long)selection_indices);
		delete selection_indices;
		return 0;
	}

	// set up data matrix
	data_matrix = gsl_matrix_alloc(4,selection_indices->size());
	transformed_data_matrix = gsl_matrix_alloc(4,selection_indices->size());
	for (j=0; j<selection_indices->size(); j++) {
		gsl_matrix_set(data_matrix,0,j,(_transformed_positions[(*selection_indices)[j]])[0]);
		gsl_matrix_set(data_matrix,1,j,(_transformed_positions[(*selection_indices)[j]])[1]);
		gsl_matrix_set(data_matrix,2,j,(_transformed_positions[(*selection_indices)[j]])[2]);
		//gsl_matrix_set(data_matrix,1,j,sphere_selection->at(j)->getPosition()[1]);
		gsl_matrix_set(data_matrix,3,j,1);
	}
	// calc transformed surfels
	gsl_blas_dgemm(CblasNoTrans, CblasNoTrans, 1.0, inv_box_transformation_matrix, data_matrix, 0.0, transformed_data_matrix);
	
	// if boxselector has shape box:
	if (boxselector->getShape() == SHAPE_BOX) {
		// remove invalid surfels
		j=0;
		counter = 0;
		scene = Scene::getInstance();
		flag_to_be_set = intToFlag(selection_index);
		
		for (j=0; j<selection_indices->size(); j++) {
			x =(float) gsl_matrix_get(transformed_data_matrix,0,j);
			y =(float) gsl_matrix_get(transformed_data_matrix,1,j);
			z =(float) gsl_matrix_get(transformed_data_matrix,2,j);
			if ( (x<=1.0)&&(y<=1.0)&&(z<=1.0)&&(x>=-1.0)&&(y>=-1.0)&&(z>=-1.0) ) {
				SurfelInterface *surfel = _all_surfels[(*selection_indices)[j]];
				scene->setSurfelFlagOn(surfel, flag_to_be_set, true, false);
				scene->setSurfelFlagOn(surfel, SurfelInterface::EMPHASISE, true, false);
				counter++;
			}
		}
	}  // if boxselector has sphere shape: 
	else if (boxselector->getShape() == SHAPE_SPHERE) {
		// remove invalid surfels
		j=0;
		counter = 0;
		scene = Scene::getInstance();
		flag_to_be_set = intToFlag(selection_index);
		for (j=0; j<selection_indices->size(); j++) {
			x =(float) gsl_matrix_get(transformed_data_matrix,0,j);
			y =(float) gsl_matrix_get(transformed_data_matrix,1,j);
			z =(float) gsl_matrix_get(transformed_data_matrix,2,j);
	
			if ( sqrt(x*x+y*y+z*z)<=1.f) {
				SurfelInterface *surfel = _all_surfels[(*selection_indices)[j]];
				scene->setSurfelFlagOn(surfel, flag_to_be_set, true, false);
				scene->setSurfelFlagOn(surfel, SurfelInterface::EMPHASISE, true, false);
				counter++;
			}
		}
	}

	gsl_permutation_free(transformation_permutation);
	gsl_matrix_free(box_transformation_matrix);
	gsl_matrix_free(inv_box_transformation_matrix);
	gsl_matrix_free(data_matrix);
	gsl_matrix_free(transformed_data_matrix);
	
	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();

	_MCdelete((unsigned long)selection_indices);
	delete selection_indices;
	QApplication::restoreOverrideCursor();
	return counter;
}

int Selector::selectAxisAlignedBox(Vector3D lowerbound, Vector3D upperbound,int selection_index, bool reset_selection) {
	QApplication::setOverrideCursor(Qt::waitCursor);
	float max_squared_distance; // the squared distance we want to query
	unsigned int position;
	int found_count;
	Scene *scene;
	SurfelInterface *surfel;
	float radius;
	Vector3D center;
	float x,y,z;
	SurfelInterface::Flag flag_to_be_set;
	std::vector<SurfelInterface*> sfls;
	std::vector<int> flags;

	if(_object_selection_changed) {
		_object_selection_changed = false;
		buildTree();
	}

	if (reset_selection) {
		_selection_tool->setActiveSelection(selection_index);
		_selection_tool->clearSelection(false);
	}
	
	if (_all_surfels_count == 0 || ((float)_valid_surfels_count / (float)_all_surfels_count) < 0.5) {
		// if we're below some surfel count, don't rebuild tree
		if (_all_surfels_count > REBUILD_BORDER)
			buildTree();
	}

	if (_valid_surfels_count == 0) {		// there aren't any surfels left
		QApplication::restoreOverrideCursor();	
		return 0;
	}
	
	// compute radius
	radius = (upperbound - lowerbound).getLength();
	radius /= 2.;
	// compute max squared distance
	max_squared_distance = radius;
	max_squared_distance *= max_squared_distance;
	// compute center
	center = (upperbound + lowerbound) * .5;

	// set the query range: 
	// NOTE: NOfNeighbours must be set to no. of surfels in order to get all neighbours.
	_kdtree->setNOfNeighbours(_all_surfels_count);
	_kdtree->queryRange(center,max_squared_distance);
	
	// query & mark found surfels!
	scene = Scene::getInstance();
	found_count = 0;
	flag_to_be_set = intToFlag(selection_index);
	

	for (unsigned int i = 0; i<_kdtree->getNOfFoundNeighbours(); i++ ) {
		position = _kdtree->getNeighbourPositionIndex(i);
		if (position >=0 && position <_all_surfels_count && _valid_surfels[position]) {
			surfel = _all_surfels[position];
			x = (_transformed_positions[position])[0];
			y = (_transformed_positions[position])[1];
			z = (_transformed_positions[position])[2];
			if ( (x<=upperbound[0])&&(y<=upperbound[1])&&(z<=upperbound[2])&&(x>=lowerbound[0])&&(y>=lowerbound[1])&&(z>=lowerbound[2]) ) {
				scene->setSurfelFlagOn(surfel, flag_to_be_set, true, false);
				scene->setSurfelFlagOn(surfel, SurfelInterface::EMPHASISE, true, false);
				found_count++;
			}
		}
	}

	scene->emitObjectModifiedSignal();
	scene->emitRenderFrameSignal();
	QApplication::restoreOverrideCursor();
	return found_count;
}

int Selector::selectAll(int selection_index) {
	QApplication::setOverrideCursor(Qt::waitCursor);
	SurfelInterface::Flag flag_to_be_set;
	std::vector<SurfelInterface*> sfls;
	std::vector<int> flags;

	if(_object_selection_changed) {
		_object_selection_changed = false;
		buildTree();
	}
	
	flag_to_be_set = intToFlag(selection_index);
	
	
	
	for (unsigned int i=0; i<_all_surfels_count; i++) 
		if (_valid_surfels[i]) {
			Scene::getInstance()->setSurfelFlagOn(_all_surfels[i], flag_to_be_set,true,false);
			Scene::getInstance()->setSurfelFlagOn(_all_surfels[i], SurfelInterface::EMPHASISE, true, false);
		}
	
		
	Scene::getInstance()->emitObjectModifiedSignal();
	Scene::getInstance()->emitRenderFrameSignal();
	
	QApplication::restoreOverrideCursor();
	return _valid_surfels_count;
}

void Selector::deselectIndex(int selection_index) {
	std::vector<SurfelInterface*> sfls;
	std::vector<int> flags;

	for (unsigned int i=0; i<_all_surfels_count; i++) 
		if (_valid_surfels[i]) {
			Scene::getInstance()->setSurfelFlagOn(_all_surfels[i], intToFlag(selection_index),false,false);
			if (_all_surfels[i]->getFlags() == SurfelInterface::EMPHASISE)
				Scene::getInstance()->setSurfelFlagOn(_all_surfels[i], SurfelInterface::EMPHASISE, false, false);
		}
	Scene::getInstance()->emitObjectModifiedSignal();
	Scene::getInstance()->emitRenderFrameSignal();
}

void Selector::deselectAll() {
	std::vector<SurfelInterface*> sfls;
	std::vector<int> flags;

	
	for (unsigned int i=0; i<_all_surfels_count; i++) 
		if (_valid_surfels[i]) {
			Scene::getInstance()->setSurfelFlagOn(_all_surfels[i], SurfelInterface::SELECTED1,false,false);
			Scene::getInstance()->setSurfelFlagOn(_all_surfels[i], SurfelInterface::SELECTED2,false,false);
			Scene::getInstance()->setSurfelFlagOn(_all_surfels[i], SurfelInterface::SELECTED3,false,false);
			Scene::getInstance()->setSurfelFlagOn(_all_surfels[i], SurfelInterface::EMPHASISE, false, false);
		}
	
	Scene::getInstance()->emitObjectModifiedSignal();
	Scene::getInstance()->emitRenderFrameSignal();
}

SelectionTool* Selector::getSelectionTool(void) {
	return _selection_tool;
}

int Selector::flagToInt(SurfelInterface::Flag convert_flag) {
	switch (convert_flag) {
		case SurfelInterface::SELECTED1:
			return 0;
		case SurfelInterface::SELECTED2:
			return 1;
		case SurfelInterface::SELECTED3:
			return 2;
		case 128:
			return 3;
		default:
			return -1;
	}
}

SurfelInterface::Flag Selector::intToFlag(int convert_int) {
	switch (convert_int) {
		case 0:
			return SurfelInterface::SELECTED1;
		case 1:
			return SurfelInterface::SELECTED2;
		case 2:
			return SurfelInterface::SELECTED3;
		case 3:
			return (SurfelInterface::Flag)128;
		default:
			return SurfelInterface::NO_FLAGS;
	}
}

KdTree* Selector::returnKdTree(void) {
	if (!_kdtree)
		buildTree();
	return _kdtree;
}

std::vector<Vector3D>* Selector::getAllPositionsWorldStdVec() {
	std::vector<Vector3D> *tmp;
	tmp = new std::vector<Vector3D>;
	_MCnew( (unsigned long)tmp, sizeof(tmp), __FILE__, __LINE__);
	
	for (unsigned i=0; i<_all_surfels_count; i++)
		if (_valid_surfels[i])
			tmp->push_back(_transformed_positions[i]);
	
	return tmp;
}

std::vector<Vector3D>* Selector::getSurfacePositionsWorldStdVec() {
	std::vector<Vector3D> *tmp;
	tmp = new std::vector<Vector3D>;
	_MCnew( (unsigned long)tmp, sizeof(tmp), __FILE__, __LINE__);
	
	for (unsigned i=0; i<_all_surfels_count; i++)
		if (_valid_surfels[i] && IsSurfelInObjList(i)) {
			tmp->push_back(_transformed_positions[i]);
		}
	
	return tmp;
}

bool Selector::getSurfacePositionNormalSurfelRadiusWorldStdVec(	
	std::vector<Vector3D> **xpositions,
	std::vector<Vector3D> **xnormals,
	std::vector<SurfelInterface*> **xsurfels,
	std::vector<float> **xradii ) {
	std::vector<Vector3D>				*positions = 0;
	std::vector<Vector3D>				*normals = 0;
	std::vector<SurfelInterface*>		*surfels = 0;
	std::vector<float>					*radii = 0;
	SurfelInterface* surfel = 0;

	positions = new std::vector<Vector3D>;
	_MCnew( (unsigned long)positions, sizeof(positions), __FILE__, __LINE__);

	normals = new std::vector<Vector3D>;
	_MCnew( (unsigned long)normals, sizeof(normals), __FILE__, __LINE__);
	
	surfels = new std::vector<SurfelInterface*>;
	_MCnew( (unsigned long)surfels, sizeof(surfels), __FILE__, __LINE__);
	
	radii = new std::vector<float>;
	_MCnew( (unsigned long)radii, sizeof(radii), __FILE__, __LINE__);

	if (!positions || !normals || !surfels || !radii) {
		if (positions) delete positions;
		if (normals) delete normals;
		if (surfels) delete surfels;
		if (radii) delete radii;
		return false;
	}
	
	for (unsigned i=0; i<_all_surfels_count; i++) {
		if (_valid_surfels[i] && IsSurfelInObjList(i)) {
			surfel = _all_surfels[i];

			positions->push_back(_transformed_positions[i]);
			normals->push_back(_transformed_normals[i]);
			surfels->push_back(surfel);
			radii->push_back(surfel->getRadius());
		}
	}

	*xpositions = positions; 
	*xnormals   =   normals;  
	*xsurfels   =   surfels;   
	*xradii     =     radii;    

	return true;
}

QVector<SurfelInterface>* Selector::getSelectionSurfels(SurfelInterface::Flag selection_flag) {
	QVector<SurfelInterface> *ret_vec;
	SurfelInterface *surfel;
	int index;

	ret_vec = new QVector<SurfelInterface>;
	_MCnew( (unsigned long)ret_vec, sizeof(ret_vec), __FILE__, __LINE__);
	ret_vec->resize(_valid_surfels_count);
	
	index = 0;
	
	for (unsigned int i=0; i<_all_surfels_count; i++) {
		if (_valid_surfels[i]) {
			surfel = _all_surfels[i];
			if (surfel->isFlagOn(selection_flag)) {
				ret_vec->insert(index, surfel);
				index++;
			}
		}
	}
	
	ret_vec->resize(index);
	return ret_vec;
}

QVector<Vector3D>* Selector::getSelectionPositions(SurfelInterface::Flag selection_flag) {
	QVector<Vector3D> *ret_vec;
	SurfelInterface *surfel;
	int index;

	ret_vec = new QVector<Vector3D>;
	_MCnew( (unsigned long)ret_vec, sizeof(ret_vec), __FILE__, __LINE__);

	ret_vec->resize(_valid_surfels_count);
	
	index = 0;
	
	for (unsigned int i=0; i<_all_surfels_count; i++) {
		if (_valid_surfels[i]) {
			surfel = _all_surfels[i];
			if (surfel->isFlagOn(selection_flag)) {
				Vector3D *a = new Vector3D(surfel->getPosition().source());
				_MCnew( (unsigned long)a, sizeof(a), __FILE__, __LINE__);

				ret_vec->insert(index, a);
				index++;
			}
		}
	}
	
	ret_vec->resize(index);
	return ret_vec;
}

QVector<Vector3D>* Selector::getSelectionPositionsWorld(SurfelInterface::Flag selection_flag) {
	QVector<Vector3D> *ret_vec;
	SurfelInterface *surfel;
	int index;

	ret_vec = new QVector<Vector3D>;
	_MCnew( (unsigned long)ret_vec, sizeof(ret_vec), __FILE__, __LINE__);

	ret_vec->resize(_valid_surfels_count);
	
	index = 0;
	
	for (unsigned int i=0; i<_all_surfels_count; i++) {
		if (_valid_surfels[i]) {
			surfel = _all_surfels[i];
			if (surfel->isFlagOn(selection_flag)) {
				Vector3D *a = new Vector3D(getSurfelWorldPosition(i).source());
				_MCnew( (unsigned long)a, sizeof(a), __FILE__, __LINE__);
				ret_vec->insert(index, a);
				index++;
			}
		}
	}
	
	ret_vec->resize(index);
	return ret_vec;
}

std::vector<Vector3D>* Selector::getSelectionPositionsWorldStdVec(SurfelInterface::Flag selection_flag) {
	std::vector<Vector3D>* ret_vec;
	SurfelInterface *surfel;
	int index;

	ret_vec = new std::vector<Vector3D>;
	_MCnew( (unsigned long)ret_vec, sizeof(ret_vec), __FILE__, __LINE__);
	ret_vec->clear();
	
	index = 0;
	
	for (unsigned int i=0; i<_all_surfels_count; i++) {
		if (_valid_surfels[i]) {
			surfel = _all_surfels[i];
			if (surfel->isFlagOn(selection_flag)) {
				ret_vec->push_back(getSurfelWorldPosition(i));
				index++;
			}
		}
	}
	
	return ret_vec;
}

QVector<unsigned int>* Selector::getSelectionIndices(SurfelInterface::Flag selection_flag) {
	QVector<unsigned int> *ret_vec;
	SurfelInterface *surfel;
	int index;

	ret_vec = new QVector<unsigned int>;
	_MCnew( (unsigned long)ret_vec, sizeof(ret_vec), __FILE__, __LINE__);

	ret_vec->resize(_valid_surfels_count);
	
	index = 0;
	
	for (unsigned int i=0; i<_all_surfels_count; i++) {
		if (_valid_surfels[i]) {
			surfel = _all_surfels[i];
			if (surfel->isFlagOn(selection_flag)) {
				unsigned int *a = new unsigned int(i);
				_MCnew( (unsigned long)a, sizeof(a), __FILE__, __LINE__);

				ret_vec->insert(index, a);
				index++;
			}
		}
	}
	
	ret_vec->resize(index);
	return ret_vec;
}

std::vector<unsigned int>* Selector::getSelectionIndicesStdVec(SurfelInterface::Flag selection_flag) {
	std::vector<unsigned int>* ret_vec;
	SurfelInterface *surfel;

	ret_vec = new std::vector<unsigned int>;
	_MCnew( (unsigned long)ret_vec, sizeof(ret_vec), __FILE__, __LINE__);

	ret_vec->clear();

	for (unsigned int i=0; i<_all_surfels_count; i++) {
		if (_valid_surfels[i]) {
			surfel = _all_surfels[i];
			if (surfel->isFlagOn(selection_flag)) {
				ret_vec->push_back(i);
			}
		}
	}
	return ret_vec;
}

bool Selector::IsSurfelInObjList(int index) {
	std::map<unsigned int, Object*>::iterator iter1;
	
	// find object to which position belongs
	iter1 = _index_bounds.lower_bound(index);
	if (iter1 == _index_bounds.end())
		return false;
	
	return IsObjInObjList((*iter1).second, mSurfaceObjectList);
}

std::vector<unsigned int>* Selector::getToProjectSurfelIndicesStdVec(SurfelInterface::Flag selection_flag, bool projectSelectionOnly) {
	std::vector<unsigned int>* ret_vec;
	SurfelInterface *surfel;

	ret_vec = new std::vector<unsigned int>;
	_MCnew( (unsigned long)ret_vec, sizeof(ret_vec), __FILE__, __LINE__);

	ret_vec->clear();

	for (unsigned int i=0; i<_all_surfels_count; i++) {
		if (_valid_surfels[i]) {
			surfel = _all_surfels[i];
			if (IsSurfelInObjList(i))
				continue;
			if (projectSelectionOnly) {
				if (surfel->isFlagOn(selection_flag)) {
					ret_vec->push_back(i);
				}
			}
			else {
				ret_vec->push_back(i);
			}
		}
	}
	return ret_vec;
}

void Selector::getSelectionSurfelsAndIndexObject(SurfelInterface::Flag selection_flag, std::vector<SurfelInterface*> &selected_surfels, std::map<unsigned int, Object*> &object_indices) {
	SurfelInterface *surfel;
	int index;
	std::map<unsigned int,Object*>::iterator index_bounds_iterator;
	
	index = 0;
	selected_surfels.clear();
	object_indices.clear();
	index_bounds_iterator = _index_bounds.begin();

	for (unsigned int i=0; i<_all_surfels_count; i++) {
		if (_valid_surfels[i]) {
			surfel = _all_surfels[i];
			if (surfel->isFlagOn(selection_flag)) {
				selected_surfels.push_back(surfel);
				index++;
				
				if ((*index_bounds_iterator).first< i) {
					//reached an index bound.
					object_indices[index] = (*index_bounds_iterator).second;
					index_bounds_iterator++;
				}
				
			}
		}
	}
	object_indices[index] = (*index_bounds_iterator).second;
}

std::vector<unsigned int>* Selector::getAllIndicesStdVec() {
	std::vector<unsigned int>* ret_vec;

	ret_vec = new std::vector<unsigned int>;
	_MCnew( (unsigned long)ret_vec, sizeof(ret_vec), __FILE__, __LINE__);

	ret_vec->clear();

	
	for (unsigned int i=0; i<_all_surfels_count; i++) 
		if (_valid_surfels[i]) 
			ret_vec->push_back(i);

	return ret_vec;
}

SurfelInterface* Selector::getSurfel(int index) {
	if (_valid_surfels[index])
		return _all_surfels[index];
	else 
		return 0;
}

int Selector::getNOfSurfels(void) {
	return _valid_surfels_count;
}

Vector3D Selector::getSurfelWorldPosition(unsigned int index) {
	if (index >=_all_surfels_count || !_valid_surfels[index])
		0;

	return _transformed_positions[index];
}

bool Selector::getSurfelWorldPosition(unsigned int index, Vector3D& return_vector) {
	if (index >=_all_surfels_count || !_valid_surfels[index])
		return false;
	return_vector = _transformed_positions[index];
	return true;
}

bool Selector::getSurfelWorldNormal(unsigned int index, Vector3D &result) {
	SurfelInterface *surfel;
	std::map<unsigned int, Object*>::iterator iter1;
	std::map<Object*, tTransformationPair,MyObjectComparator>::iterator iter2;
	
	if (index >=_all_surfels_count || _valid_surfels[index] == false)
		return false;

	surfel = _all_surfels[index];

	// find object to which position belongs
	iter1 = _index_bounds.lower_bound(index);
	if (iter1 == _index_bounds.end()) 
		return false;
	
	iter2 =	_selected_objects.find( (*iter1).second );
	if (iter2 == _selected_objects.end())
		return false;

	MtrMultVector3D(((*iter2).second.transformation), surfel->getNormal(), &result);
	return true;	
}

void Selector::setSurfelWorldPosition(unsigned int index, Vector3D newposition, bool repaintRenderer) {
	Vector3D result;
	std::map<unsigned int, Object*>::iterator iter1;
	std::map<Object*, tTransformationPair,MyObjectComparator>::iterator iter2;
	
	
	if (!_valid_surfels || index >=_all_surfels_count || _valid_surfels[index] == false)
		return;

	// find object to which position belongs
	iter1 = _index_bounds.lower_bound(index);
	if (iter1 == _index_bounds.end()) 
		return;
	
	iter2 =	_selected_objects.find( (*iter1).second );
	if (iter2 == _selected_objects.end())
		return;

	MtrMultVector3DHom(((*iter2).second.inverse_transformation), newposition, &result);
	Scene::getInstance()->setSurfelPosition(_all_surfels[index], result, repaintRenderer);
	_transformed_positions[index] = newposition;
}


void Selector::setSurfelWorldNormal(unsigned int index, Vector3D newnormal, bool repaintRenderer) {
	Vector3D result;
	std::map<unsigned int, Object*>::iterator iter1;
	std::map<Object*, tTransformationPair,MyObjectComparator>::iterator iter2;
	
	
	if (!_valid_surfels || index >=_all_surfels_count || _valid_surfels[index] == false)
		return;

	// find object to which position belongs
	iter1 = _index_bounds.lower_bound(index);
	if (iter1 == _index_bounds.end()) 
		return;
	
	iter2 =	_selected_objects.find( (*iter1).second );
	if (iter2 == _selected_objects.end())
		return;

	MtrMultVector3D(((*iter2).second.inverse_transformation), newnormal, &result);
	Scene::getInstance()->setSurfelNormal(_all_surfels[index], result, repaintRenderer);
	_transformed_normals[index] = result;
}

void Selector::addAllObjectsToSelection() {
	Scene *scene;
	Object *object;

	removeAllObjects();

	scene = Scene::getInstance();
	for (object = scene->getFirstObject(); object != 0; object = scene->getNextObject()) 
		addObjectToSelection(object);

}

void Selector::addObjectToSelection(Object *obj) {
	tTransformationPair ttp;
	MyDataTypes::TransformationMatrix16f trafo;
	MyDataTypes::TransformationMatrix16f invtrafo;
	std::map<Object*, tTransformationPair,MyObjectComparator>::iterator iter;
	
	iter = _selected_objects.find(obj);
	if (iter == _selected_objects.end()) {
		obj->getTransformationMatrix(trafo);
		if (!MtrInverse4x4f(trafo, invtrafo))
			qDebug("Singular Transformation Matrix!!!");
		memcpy(ttp.transformation, trafo, 16*sizeof(float));
		memcpy(ttp.inverse_transformation, invtrafo, 16*sizeof(float));
		_selected_objects[obj] = ttp;
		_object_selection_changed = true;
	}
}

void Selector::addObjectToSurfaceDef(Object *obj) {
	mSurfaceObjectList.push_back(obj);
}

void Selector::removeAllObjects() {
	
	_selected_objects.clear();
	mSurfaceObjectList.clear();
	_object_selection_changed = true;
}

void Selector::removeObjectFromSelection(Object *obj) {
	std::map<Object*, tTransformationPair,MyObjectComparator>::iterator iter;
	
	iter = _selected_objects.find(obj);
	if (iter != _selected_objects.end()) {
		_selected_objects.erase(iter);
		_object_selection_changed = true;
	}
	
}

Object* Selector::getFirstSelectedObject() {
	_current_object_iterator = _selected_objects.begin();
	if (_current_object_iterator == _selected_objects.end())
		return 0;
	else
		return (*_current_object_iterator).first;
}

Object* Selector::getNextSelectedObject() {
	_current_object_iterator ++;
	if (_current_object_iterator == _selected_objects.end())
		return 0;
	else
		return (*_current_object_iterator).first;
}

#ifdef _WIN32
//	#pragma warning(pop)
#endif
// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
